//**************************************************************************
//    start_explain.C
//
//    A simple program to hook up to the coral server, and then run the
//    Explain tcl GUI.
//
//    This program expects its initial input from a pipe hooked to stdin.
//
//**************************************************************************

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/wait.h>
#include <sys/param.h>

    // This is a hack, but otherwise we need the entire coral library.
#define    _NO_CORAL

#include "CoralSock.h"

//************************************************************************
// CoralSock Constructor -
//    Opens a TCP socket (an Internet stream socket) as a server, binding
//    to the port specified. If the port is zero, this connection is
//    "local" and the socket is connected to stdin/stdout instead. 
//
//************************************************************************
CoralSock::CoralSock(unsigned short Port)
{
    struct sockaddr_in    serv_addr;
    char hostnme[HOSTNAME_LENGTH];

    gethostname(hostnme, HOSTNAME_LENGTH);
    strcpy(host,hostnme);
    port = Port;

    if (Port){

          // Normal server setup
        if ( (socketfd = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
#ifdef    _NO_CORAL
            fprintf(stderr,"server: can't open stream socket\n");
#else
            fprintf(Cor_exEnv->error_file,"server: can't open stream socket\n");
#endif
            socketfd = -1;
            return;
        }
        
          // Bind the local address so that the client can send to this socket.
        memset((void *) &serv_addr, 0, sizeof(serv_addr));
        serv_addr.sin_family      = AF_INET;
        serv_addr.sin_addr.s_addr = htonl(INADDR_ANY);
        serv_addr.sin_port        = htons(Port);
        
        if (bind(socketfd, (struct sockaddr *) &serv_addr,
                                        sizeof(serv_addr)) < 0){
#ifdef    _NO_CORAL
            fprintf(stderr,"server:can't bind local address\n");
#else
            fprintf(Cor_exEnv->error_file,"server:can't bind local address\n");
#endif
            socketfd = -1;
            return;
        }
    
          // listen() specifies the queue size for pending connections
        listen(socketfd, 3);

    } else {

          // This is a "Local" Connection (not in server mode)
        socketfd = 0;
        in = stdin;
        out = stdout;
    }

    socketType = ServerType;
}

//************************************************************************
// CoralSock Constructor -
//    Opens a socket to the coral server specified the envirornment
//    variables CORALSERVER and CORALPORT. If these variables aren't
//    set the defaults are used.
//
//************************************************************************
CoralSock::CoralSock()
{
    unsigned short c_port;

      // Find a coral server to use (if its null, thats okay)
    char *coral_server = getenv("CORALSERVER");

      // The port (usually will just be the default)
    char *coral_port = getenv("CORALPORT");

    if(!coral_port)
        c_port = (unsigned short) CORAL_PORT;
    else
        c_port = (unsigned short) atoi(coral_port);

    open_client(this, coral_server, c_port);
}

//************************************************************************
// open_client() -
//    A private function for use by the CoralSock client socket
//    constructors.  Sets up the CoralSock object passed in to be a
//    connection to the Coral Server specified by HostName and Port.
//
//************************************************************************
CoralSock *CoralSock::open_client(CoralSock *cs, char *HostName,
                    unsigned short Port)
{
    struct sockaddr_in    serv_addr;
    struct hostent *host;
    char hostnme[HOSTNAME_LENGTH];

    // Fill in the structure "serv_addr" with the address of the
    // coral server that we want to connect with.

    if( HostName && *HostName ){
        strcpy(cs->host,HostName);
    } else {
        gethostname(hostnme, HOSTNAME_LENGTH);
        strcpy(cs->host,hostnme);
    } 

    cs->port = Port; 
    host = gethostbyname(cs->host);
    
    serv_addr.sin_family      = AF_INET;
    serv_addr.sin_addr.s_addr = ((struct in_addr*)(host->h_addr))->s_addr;
    serv_addr.sin_port        = htons(Port);

    if ((cs->socketfd = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
#ifdef    _NO_CORAL
        fprintf(stderr,"client: can't open stream socket\n");
#else
        fprintf(Cor_exEnv->error_file,"client: can't open stream socket\n");
#endif
        cs->socketfd = -1;
        return((CoralSock *) NULL);
    }
    if (connect(cs->socketfd,(struct sockaddr*)&serv_addr,
                sizeof(serv_addr)) < 0){
#ifdef    _NO_CORAL
        fprintf(stderr,"client: can't connect to server\n");
#else
        fprintf(Cor_exEnv->error_file,"client: can't connect to server\n");
#endif
        socketfd = -1;
        return((CoralSock *) NULL);
    }

    cs->socketType = ClientType;
    return(cs);
}

//************************************************************************
// AcceptConnection() -
//    Accepts a connection on this server CoralSock, returning a new
//    CoralSock object to use to communicate with the client.
//
//    Return: A CoralSock object for talking to the client, or NULL
//        if an error occured.
//
//************************************************************************
CoralSock *CoralSock::AcceptConnection()
{
    if ((socketfd < 0) || (socketType != ServerType))
        return((CoralSock *) NULL);

      // This is a "Local" connection (not in server mode)
    if (socketfd == 0)
        return(this);

    struct sockaddr Addr;
    int AddrLen = sizeof(Addr);
    
    int sd = accept(socketfd, &Addr, &AddrLen);
    if (sd < 0) {
        perror("CoralSock: accept");
        return((CoralSock *) NULL);
    }

    return new CoralSock(sd, ClientType);
}

//************************************************************************
// RefuseConnection() -
//    Refuses a connection on this server CoralSock object. This function
//    is used when the Coral Server is already at the maximum number of
//    connections when a connection request is recieved.  If something
//    isn't done, they will bug us every time we call select(). Instead,
//    accept the connection, then tell them to get lost.
//
//************************************************************************
void CoralSock::RefuseConnection()
{
    if ((socketfd < 0) || (socketType != ServerType))
        return;

      // This is a "Local" connection (not in server mode)
    if (socketfd == 0)
        return;

    struct sockaddr Addr;
    int AddrLen = sizeof(Addr);
   
      // Temporially accepting, so we can tell them to get lost. 
    int sd = accept(socketfd, &Addr, &AddrLen);
    if (sd < 0) {
        perror("CoralSock: accept");
        return;
    }

    char *mess = "Connection refused by Coral Server.\n";
    write(sd,mess,strlen(mess));

      // Disconnect
    shutdown(sd,2);
    close(sd);
}

//************************************************************************
// InFilePtr() - 
//    Returns a file pointer to be used to read from this CoralSock socket.
//    Only one input file pointer is created for each socket, and given
//    out when someone wants it. Because of this, the file pointer
//    given by InFilePtr() should NOT be closed.
//
//    Return: A file pointer for reading from this CoralSock socket, or
//        NULL if an error occurs.
//
//************************************************************************
FILE *CoralSock::InFilePtr()
{
    if (in)
        return(in);

    if (socketfd < 0)
        return((FILE *) NULL);

    return(in = fdopen(socketfd, "r"));
}

//************************************************************************
//  OutFilePtr() -
//    Returns a file pointer to be used to write to this CoralSock socket.
//    Only one output file pointer is created for each socket, and given
//    out when someone wants it. Because of this, the file pointer
//    given by OutFilePtr() should NOT be closed. The file pointer is
//    set to be line-buffered, so it should not require flushing all
//    the time.
//
//    Return: A file pointer for writing to this CoralSock socket, or
//        NULL if an error occurs.
//
//************************************************************************
FILE *CoralSock::OutFilePtr()
{
    if (out)
        return(out);

    if (socketfd < 0)
        return((FILE *) NULL);

    if (! (out = fdopen(socketfd, "w")))
        return((FILE *) NULL);

      // Set socket file pointer to be line buffered (flushs on a newline)
    if (setvbuf(out, NULL, _IOLBF, BUFSIZ) != 0){
        fclose(out);
        out = (FILE *) NULL;
    }

    return(out);
}

//**************************************************************************
void die(int k)
{
    fprintf(stderr,
        "\n%d)Could not obtain start up info from stdin, exiting.\n\n",k);
    exit(1);
}

//**************************************************************************

int main(int argc, char **argv)
{
    int stat, pid;
    char server[HOSTNAME_LENGTH], password[10], buff[500];
    int conn_id, user_id;
    unsigned short port;

    if (argc < 3) {
        fprintf(stderr, "argc is: %d, argv[0] is: %s\n", argc, argv[0]);
        die(1);
    }

    strcpy(server,argv[1]);
    sscanf(argv[2],"%hu", &port);

      // Get info from stdin (will be a pipe)
    if (!fgets(buff, 500, stdin))
        die(2);
    sscanf(buff,"%d %d", &conn_id, &user_id);

    if (!fgets(password, 10, stdin))
        die(3);

//    fprintf(stdout, "Server: %s  Port: %hu\n", server, port);
//    fprintf(stdout, "Password: '%s'  Connection: (%d,%d)\n",password,
//        conn_id, user_id);

      // Must detach from the process that spawned us
    if ((pid = fork()) < 0){
        die(4);
    } else if (pid > 0){
          // parent must return so that tkcoral continues.
        return(0);
    }

      // Open a port to recieve explain info
    CoralSock *sock = new CoralSock(server, port);

    if (! sock->Valid()) {
        fprintf(stderr,"Explain: Couldn't connect to coral server (%s,%hu).\n",
            server, port);
        return(1);
    }

      // Tell coral server the connection id, user id, and password
    sprintf(buff,"%d %d %s", conn_id, user_id, password);
    write(sock->socketfd, buff, strlen(buff));

    dup2(sock->socketfd,0);
    //dup2(sock->socketfd,1);

      // Now start up the Explain tcl GUI and hook it in.

    char *croot = getenv("CORALROOT");
    if (!croot){
        fputs("You must set the envirornment variable CORALROOT.\n",stderr);
        return(1);
    }

    if ((pid = fork()) < 0){
        fprintf(stderr,"start_explain: Can't fork\n");
        return(1);
    }

    if (pid == 0){
            // child
        char *path = new char[MAXPATHLEN];
        char *path2 = new char[MAXPATHLEN];

        sprintf(path,"%s/bin/coralsh",croot);
        sprintf(path2,"%s/bin/Explain",croot);

        execl(path, "coralsh", "-f", path2, (char *) 0);
          // NOTREACHED
        delete [] path;
        delete [] path2;
        return(0);
    } else {
          // Hang around so coral server doesn't hang up
        wait(&stat);
    }

    delete sock;
    return(0);
}

//**************************************************************************
