// *************************************************************************
//    CArg.C
// *************************************************************************

#include <stddef.h>
#include <ctype.h>

#include "CArg.h"

#include "Globals.h"
#include "Relation.h"
#include "CoralError.h"
#include "CoralStream.h"
#include "BuiltinSymbols.h"
#include "Operators.h"
#include "StorageRelation.h"
#include "HashSimpleRelation.h"
#include "Debug.h"

#define  _CARG_C
#include "CArg.i"
#undef   _CARG_C

static long print_width = -1;
static long print_precision = -1;
static char print_double_format = 'g';
static char print_int_format = 'd';

Cor_CArg    *Cor_TheDontCareArg = NULL;

// *************************************************************************
Cor_CArg::Cor_CArg(const char *name)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
#ifdef    DO_PURE
    _doub  = 0;       // To stop purify from seeing UMR everywhere
    _hash = 0;
#endif

    unsigned int len = strlen(name) + 1;

    if (len > sizeof(double)) {

        my_arg_kind = SYMBOL_ARG;

          // insert() will allocate space if necessary
        sym._symbol = Cor_string_ref_table->insert((char*)name,
                                        &(sym._str_ref_cnt),1,&_hash,0);
    } else {
        my_arg_kind = TINY_SYM_ARG;
        _hash = Cor_StrHash(name);
        memcpy(_symbuf,name,len);
    }
}

// *************************************************************************
Cor_CArg::Cor_CArg(Cor_Relation *relation)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
#ifdef    DO_PURE
    _doub  = 0;       // To stop purify from seeing UMR everywhere
    _hash = 0;
#endif

    my_arg_kind = REL_ARG;
    _rel = relation;
    _rel->carg_count++;
}

// *************************************************************************
// Constructors for Lists
Cor_CArg::Cor_CArg(Cor_CArgStack *stk)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    my_arg_kind = LIST_ARG;

    list._items = stk;
    list._num_items = list._items->count();

    ASSERT(list._num_items);

    _hash = UnknownHashValue;
}

// *************************************************************************
// We take possesion of car/cdr
Cor_CArg::Cor_CArg(Cor_CArg *car, Cor_CArg *cdr)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_CArg       *my_cdr;

    ASSERT(car);

    if (cdr)
        my_cdr = cdr;
    else
        my_cdr = new Cor_CArg(NIL_ARG);
    
    my_arg_kind = LIST_ARG;

    if (my_cdr->is_cons()) {
        list._items = my_cdr->list._items;
        list._items->push(car);
        list._num_items = list._items->count();
    } else {
        list._items = new Cor_CArgStack();
        list._items->push(my_cdr);
        list._items->push(car);
        list._num_items = 2;     
    }

    _hash = UnknownHashValue;

    ASSERT(list._num_items > 1);
}

// *************************************************************************
Cor_CArg *Cor_CArg::car() const
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    switch (my_arg_kind) {

    default:
        ASSERT(0);
        return(NULL);

    case NIL_ARG:
        return((Cor_CArg *) this);

    case LIST_ARG:
        return(list._items->get_ith(list._num_items-1));
    }
}

Cor_CArg *Cor_CArg::cdr() const
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    switch (my_arg_kind) {

    default:
        ASSERT(0);
        return(NULL);

    case NIL_ARG:
        return(((Cor_CArg*) this)->dup());

    case LIST_ARG:
        if (list._num_items == 2)
            return(list._items->get_ith(0)->dup());
        else {
            Cor_CArg *ret = ((Cor_CArg*) this)->dup();
            delete ret->list._items->pop();
            (ret->list._num_items)--;
            ASSERT(ret->num_items() > 1);
            return(ret);
        }
    }
}

void Cor_CArg::get_cdr(Cor_CArg& cdr) const
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    switch (my_arg_kind) {

    default:
        ASSERT(0);
        cdr.init();
        break;

    case NIL_ARG:
        cdr = *this;
        break;

    case LIST_ARG:
        if (list._num_items == 2) {
              // Result is no longer a list
            cdr = *(list._items->get_ith(0));
        } else {
            cdr = *this;
            (cdr.list._num_items)--;
            ASSERT(cdr.list._num_items >= 2);
            cdr._hash = UnknownHashValue;
        }
        break;
    }
}

// *************************************************************************
Cor_CArg::Cor_CArg(const Cor_CArg &that)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_CArg *that2;
    my_arg_kind = that.my_arg_kind;

#ifdef    DO_PURE
    _doub  = 0;       // To stop purify from seeing UMR everywhere
    _hash = 0;
#endif

    switch (my_arg_kind) {

    default:
#ifdef    DEBUG
        TELL;
        cerr << "Unknown kind " << kind_str() << endl;
#endif
    case EMPTY_ARG:
        init();
        break;

    case INT_ARG:
        _int = that._int;
        break;

    case DOUBLE_ARG:
        _doub = that._doub;
        break;

    case SYMBOL_ARG: {
            _hash = that._hash;
            sym._symbol = that.sym._symbol;
            sym._str_ref_cnt = that.sym._str_ref_cnt;
            *(sym._str_ref_cnt) += 1;
        }
        break;

    case TINY_SYM_ARG: {
            _hash = that._hash;
            memcpy(_symbuf, that._symbuf, sizeof(double));
        }
        break;

    case CHUNK_ARG:
        _chunk = that._chunk;
        break;

    case FUNCT_ARG:
        that2 = (Cor_CArg *) &that;
        func._functor = that2->func._functor->dup();
        func._args = that2->func._args->dup();
        _hash = that2->hash();
        break;

    case VAR_ARG:
        _var_num = that._var_num;
        break;

    case REL_ARG:
        _rel = that._rel;
        _rel->carg_count++;
        break;

    case DONT_CARE_ARG:
    case PARAMETER_ARG:
        break;

      // Lists
    case LIST_ARG:
        that2 = (Cor_CArg*) &that;
        list._items = that2->list._items->dup();
        list._num_items = that2->list._num_items;
        _hash = that2->hash();
        break;

    case NIL_ARG:
        break;

      // Arithmetic Expressions
    case ADD_OP_ARG:
    case SUB_OP_ARG:
    case MULT_OP_ARG:
    case DIV_OP_ARG:
    case IDIV_OP_ARG:
    case MOD_OP_ARG:

    case LSHFT_OP_ARG:
    case RSHFT_OP_ARG:

    case BT_AND_OP_ARG:
    case BT_OR_OP_ARG:
    case BT_NOT_OP_ARG:
    case BT_XOR_OP_ARG:
          // Deep copy
        expr._larg = that.larg()->dup();
        expr._rarg = that.rarg()->dup();
        break;

      // The various types of Grouping Args
    case SUM_GRPING_ARG:
    case PROD_GRPING_ARG:
    case CNT_GRPING_ARG:
    case AVG_GRPING_ARG:

    case SUM_DGRPING_ARG:
    case PROD_DGRPING_ARG:
    case CNT_DGRPING_ARG:
    case AVG_DGRPING_ARG:

    case MIN_GRPING_ARG:
    case MAX_GRPING_ARG:
    case SET_GRPING_ARG:
    case ANY_GRPING_ARG:
    case NO_AGG_GRPING_ARG:
        _chk_subsum      = that._chk_subsum;
        grp._nested_args = that.grp._nested_args->dup();
        grp._seq_num     = grp._seq_num;
        break;

    case AGG_RESULT_ARG:
          // Deep copy
        _agg_num = that._agg_num->dup();
        agg._agg_count = that.agg._agg_count;

        agg._dist_rel = that.agg._dist_rel;
        if (agg._dist_rel)
            (agg._dist_rel->carg_count)++;
        break;
    }
}

// *************************************************************************
int Cor_CArg::functor_equals(Cor_CArg *carg)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    // The caller should have checked this already
    ASSERT(is_functor());

    if (this == carg)
        return(1);

    if (!carg->is_functor())
        return(0);

    if (((Cor_CArg*) this)->arity() != carg->arity())
        return(0);

    if ( (hash() != carg->hash())
            && (Cor_isGroundHashValue(_hash))
            && (Cor_isGroundHashValue(carg->_hash)) )
        return(0);

    if (! nm_equals(*carg) )
        return(0);

    for (int i=0; i < arity(); i++) {
        if (! args()[i].equals(&(carg->args()[i])) )
            return(0);
    }

    return(1);
}

int Cor_CArg::list_equals(Cor_CArg *carg)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    // The caller should have checked this already
    ASSERT(is_cons());
    ASSERT(carg->is_cons());

    if (!carg->is_cons())
        return(0);

    if ( (hash() != carg->hash())
            && (Cor_isGroundHashValue(_hash))
            && (Cor_isGroundHashValue(carg->_hash)) ) {
        return(0);
    }

    ASSERT(num_items());
    ASSERT(carg->num_items());

    if (num_items() != carg->num_items())
        return(0);

    int j;
    Cor_CArg *item1, *item2;
    for (j=num_items()-1; j >= 0; j--) {
        item1 = items()->get_ith(j);
        item2 = carg->items()->get_ith(j);
        if (!item1->equals(item2))
            return(0);
    }

    return(1);
}

// *************************************************************************
int Cor_CArg::equals(Cor_CArg *carg)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
      // This should speed comparisons between reference-counted items
    if (this == carg)
        return(1);

    if (!this || !carg || !same_kind(carg))
        return(0);

    switch (my_arg_kind) {

    default:
#ifdef    DEBUG
        TELL;
        cerr << "Unknown kind " << kind_str() << endl;
#endif
        CORAL_error(COR_FATAL_ERROR, "", "equals", __FILE__, __LINE__);
    case EMPTY_ARG:
        return(0);

    case INT_ARG:
        return(_int == carg->_int);

    case DOUBLE_ARG: {
            double diff = _doub - carg->_doub;
            if (diff < 0)
                diff = -diff;
            return(diff < .000001);
        }

    case SYMBOL_ARG:
        if (sym._symbol == carg->sym._symbol)
            return(1); 
        break;

    case TINY_SYM_ARG:
        if (_hash != carg->_hash)
            return(0);

        if (strcmp(_symbuf, carg->_symbuf) == 0)
            return(1);
        break;

    case CHUNK_ARG:
        return(_chunk == carg->_chunk);

    case FUNCT_ARG:
        return(functor_equals(carg));

    case VAR_ARG:
        return(_var_num == carg->_var_num);

    case REL_ARG:
        return(rel()->equals(carg->rel()));
        break;

    case DONT_CARE_ARG:
        if (carg->Kind() == DONT_CARE_ARG)
            return(1);
        break;

    case PARAMETER_ARG:
        return(1);

      // Lists
    case NIL_ARG:
        if (carg->Kind() == NIL_ARG)
            return(1);
        break;

    case LIST_ARG:
        return(list_equals(carg));

      // Arithmetic Expressions
    case ADD_OP_ARG:
    case SUB_OP_ARG:
    case MULT_OP_ARG:
    case DIV_OP_ARG:
    case IDIV_OP_ARG:
    case MOD_OP_ARG:

    case LSHFT_OP_ARG:
    case RSHFT_OP_ARG:

    case BT_AND_OP_ARG:
    case BT_OR_OP_ARG:
    case BT_NOT_OP_ARG:
    case BT_XOR_OP_ARG:
        return(   larg()->equals(carg->larg())
               && rarg()->equals(carg->rarg()) );
        break;

      // The various types of Grouping Args
    case SUM_GRPING_ARG:
    case PROD_GRPING_ARG:
    case CNT_GRPING_ARG:
    case AVG_GRPING_ARG:

    case SUM_DGRPING_ARG:
    case PROD_DGRPING_ARG:
    case CNT_DGRPING_ARG:
    case AVG_DGRPING_ARG:

    case MIN_GRPING_ARG:
    case MAX_GRPING_ARG:
    case SET_GRPING_ARG:
    case ANY_GRPING_ARG:
    case NO_AGG_GRPING_ARG:
        return(0);

    case AGG_RESULT_ARG:
        TELL;
        cerr << "Need code here yet.\n";
        break;
    }

    return(0);
}

// *************************************************************************
// compares _any_ two args, return <0, =0, or >0
// depending on relation being object and "that".
// The result is arbitrary (but consistent),
// when the types are incompatible.
// Equality only means "approximately", as this function is intended for
// sorting, even if no meaningful ordering is possible.
// *************************************************************************
int Cor_CArg::compare(Cor_CArg *that) const
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
      // This should speed comparisons between reference-counted items
    if (this == that)
        return(0);

    if (! same_kind(that)) {
        if (this->my_arg_kind < that->my_arg_kind)
            return(-1);
        else
            return(1);
    }

    switch (my_arg_kind) {

    default:
#ifdef    DEBUG
        TELL;
        cerr << "Unknown kind " << kind_str() << endl;
#endif
        CORAL_error(COR_FATAL_ERROR, "", "compare", __FILE__, __LINE__);
    case EMPTY_ARG:
        return(0);

    case INT_ARG:
    case DOUBLE_ARG:
        return(compare_num(*that));

    case SYMBOL_ARG:
        if (that->my_arg_kind == SYMBOL_ARG)
            return( strcmp(sym._symbol, that->sym._symbol) );
        else
            return( strcmp(sym._symbol, that->_symbuf) );

    case TINY_SYM_ARG:
        if (that->my_arg_kind == TINY_SYM_ARG)
            return( strcmp(_symbuf, that->_symbuf) );
        else
            return( strcmp(_symbuf, that->sym._symbol) );

    case CHUNK_ARG:
        if (_chunk == that->_chunk)
            return(0);
        if (_chunk < that->_chunk)
            return(-1);
        else
            return(1);

    case FUNCT_ARG:
        return(strcmp(functor(), that->functor()));

    case VAR_ARG:
        if (_var_num == that->_var_num)
            return(0);
        if (_var_num < that->_var_num)
            return(-1);
        else
            return(1);

    case REL_ARG:
    case DONT_CARE_ARG:
    case PARAMETER_ARG:
        break;

      // Lists
    case NIL_ARG:
    case LIST_ARG:
        break;

      // Arithmetic Expressions
    case ADD_OP_ARG:
    case SUB_OP_ARG:
    case MULT_OP_ARG:
    case DIV_OP_ARG:
    case IDIV_OP_ARG:
    case MOD_OP_ARG:

    case LSHFT_OP_ARG:
    case RSHFT_OP_ARG:

    case BT_AND_OP_ARG:
    case BT_OR_OP_ARG:
    case BT_NOT_OP_ARG:
    case BT_XOR_OP_ARG:
        break;

      // The various types of Grouping Args
    case SUM_GRPING_ARG:
    case PROD_GRPING_ARG:
    case CNT_GRPING_ARG:
    case AVG_GRPING_ARG:

    case SUM_DGRPING_ARG:
    case PROD_DGRPING_ARG:
    case CNT_DGRPING_ARG:
    case AVG_DGRPING_ARG:

    case MIN_GRPING_ARG:
    case MAX_GRPING_ARG:
    case SET_GRPING_ARG:
    case ANY_GRPING_ARG:
    case NO_AGG_GRPING_ARG:

    case AGG_RESULT_ARG:
        break;
    }

    return(0);
}

// *************************************************************************
Cor_CArg *Cor_CArg::normalize_list()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    // Clean up the inside of the list, destroy what we don't need

    // Shrink the list to it's "real" size
    while (num_items() < items()->count()) {
        delete items()->pop();
    }

    ASSERT(is_cons());

    Cor_CArg *zarg = items()->get_ith(0);

    if (!zarg->is_cons())
        return(this);

    items()->clr_ith(0);
    zarg = zarg->normalize_list();

    int j, ni = num_items();
    zarg->incr_num_items(ni-1);

    for (j=1; j < ni; j++) {
        Cor_CArg *tmp = items()->get_ith(j);
        items()->clr_ith(j);
        zarg->items()->push(tmp);
    }

    delete this;

    return(zarg);
}

// *************************************************************************
// Determine if this item needs quotes when its printed
int  Cor_CArg::need_quotes(char *str) const
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (isupper(*str) || isdigit(*str))
        return(1);

      // Empty strings must be quoted
    if (!(*str))
        return(1);

    while (*str) {

        switch (*str) {
    
        case ' ':
        case '\t':
        case '\n':
        case '\f':
        case '\a':
        case '\r':
        case '\v':
        case '\b':
        case '\\':
        case '.':
        case ',':
        case '(': case ')':
        case '[': case ']':
        case '{': case '}':
        case '|':
            return(1);

        case '"':
            cerr << "quote inside a string, printing isn't going to work.\n";
            return(-1);
        }

        str++;
    }

    return(0);
}

// *************************************************************************
//  Functors
Cor_CArg::Cor_CArg(Cor_CArg& functor, int arity)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    func._functor = new Cor_CArg(functor);

    my_arg_kind = FUNCT_ARG;
    func._args = new Cor_ArgList(arity);

    _hash = UnknownHashValue;
}

int Cor_CArg::arity() {

    switch (my_arg_kind) {
    
    default:
        cerr << "Kind is " << kind_str() << endl;
        ASSERT(0);
        return(0);
    
    case FUNCT_ARG:    
        return(func._args->arity());
    }
}

// *************************************************************************
HashVal Cor_CArg::functor_hash(Cor_BindEnv *env) 
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    HashVal my_hash = UnknownHashValue;

    switch (_hash) {

    default:
        return(_hash);

    case VarHashValue:
            // Functor has variables, and they are definitely not bound.
        if (env == NULL)
            return(VarHashValue);

        // Fall through
    case UnknownHashValue: {

        HashVal arg_hash = 0;

        my_hash = Cor_StrHash(functor());

        // Give all arguments a chance to be hashed?
        FOR_EACH_ARG(farg, args()) {

            arg_hash = farg->hash(env);

            if (arg_hash == VarHashValue)
                return(_hash = VarHashValue);

            my_hash = (my_hash << 1) ^ (my_hash >> 20) ^ arg_hash;

        } END_EACH_ARG;

          // To ensure we did not hash to UnknownHashValue or VarHashValue
        if (my_hash >= UnknownHashValue) {
            my_hash = 1;
#ifdef    DEBUG
            // Just cause I'm curious.
    cerr<<"Extreemely improbable hashing problem narrowly avoided (functor).\n";
#endif
        }
    }
    }

    return(my_hash);
}

// *************************************************************************
HashVal Cor_CArg::list_hash(Cor_BindEnv *env) 
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    HashVal my_hash = UnknownHashValue;

    ASSERT(is_cons());

    switch (_hash) {

    default:
        return(_hash);

    case VarHashValue:
          // List has variables, and they are definitely not bound.
        if (env == NULL)
            return(VarHashValue);

    case UnknownHashValue: {

        int j, lim;
        Cor_CArg *item;
        HashVal itm_hash;
        my_hash = 0;

        lim = num_items();

        for (j=0; j < lim; j++) {
            item = items()->get_ith(j);
            itm_hash = item->hash(env);

            ASSERT(itm_hash != UnknownHashValue);

            if (itm_hash == VarHashValue)
                return(VarHashValue);
            if (j)
                my_hash = ((itm_hash << 3) ^ (itm_hash >> 10) ^ my_hash) +3;
            else
                my_hash = itm_hash;
        }

          // To ensure we did not hash to UnknownHashValue or VarHashValue
        ASSERT(my_hash < UnknownHashValue);
        }
    }

    return(my_hash);
}

// *************************************************************************
HashVal Cor_CArg::hash(Cor_BindEnv *context) const
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif

    switch (my_arg_kind) {

    default:
#ifdef    DEBUG
        TELL;
        cerr << "Unknown kind in hash(): " << kind_str() << endl;
#endif
        ASSERT(0);
        break;

    case EMPTY_ARG:
#ifdef    DEBUG
        TELL;
        cerr << "Just hashed an empty arg (but that might be okay).\n";
#endif
        return(0);

    case INT_ARG:
    case DOUBLE_ARG:
        return(Cor_IntToHash(_int));

    case SYMBOL_ARG:
    case TINY_SYM_ARG:
        return(_hash);

    case CHUNK_ARG:
        return((HashVal) _chunk);

    case FUNCT_ARG: {
        Cor_CArg *this2 = (Cor_CArg *) this;
        if (context)
            return(this2->functor_hash(context));
        else
            return(this2->_hash = this2->functor_hash(context));
        }

    case VAR_ARG: {
        if (context == NULL)
            return(VarHashValue);

        return( context->hash_var(_var_num) );
        }

    case REL_ARG:
        return(rel()->hash(context));
        break;

    case DONT_CARE_ARG:
        return(VarHashValue);

    case PARAMETER_ARG:
        return(0);

    case NIL_ARG:
        return(0);

      // Lists
    case LIST_ARG: {
        Cor_CArg *this2 = (Cor_CArg *) this;
        if (context)
            return(this2->list_hash(context));
        else
            return(this2->_hash = this2->list_hash(context));
        }

      // Arithmetic Expressions
    case ADD_OP_ARG:
    case SUB_OP_ARG:
    case MULT_OP_ARG:
    case DIV_OP_ARG:
    case IDIV_OP_ARG:
    case MOD_OP_ARG:

    case LSHFT_OP_ARG:
    case RSHFT_OP_ARG:

    case BT_AND_OP_ARG:
    case BT_OR_OP_ARG:
    case BT_NOT_OP_ARG:
    case BT_XOR_OP_ARG: {

        Cor_TermLink *renamed_vars = NULL;
        Cor_CArg a1;

        Cor_CArg *this2 = (Cor_CArg *) this;
        this2->simplify(a1, context, renamed_vars, context, NULL);

        if (!a1.is_empty())
            return(a1.hash(context));
        else
            return(VarHashValue);
        }
        break;

      // The various types of Grouping Args
    case SUM_GRPING_ARG:
    case PROD_GRPING_ARG:
    case CNT_GRPING_ARG:
    case AVG_GRPING_ARG:

    case SUM_DGRPING_ARG:
    case PROD_DGRPING_ARG:
    case CNT_DGRPING_ARG:
    case AVG_DGRPING_ARG:

    case MIN_GRPING_ARG:
    case MAX_GRPING_ARG:
    case SET_GRPING_ARG:
    case ANY_GRPING_ARG:
    case NO_AGG_GRPING_ARG:
        return(VarHashValue);

    case AGG_RESULT_ARG:
        return(VarHashValue);
    }

    cerr << "An item (of kind: " << kind_str() << " has no hash() function.\n";
    ASSERT(0);

    return(UnknownHashValue);
}

// *************************************************************************
void Cor_CArg::sprint_grping(char *str, int *pos, Cor_BindEnv *env,
        char *start, char *end) const
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_ArgList  arg_ptr;

    sprintf(str + *pos, start);
    *pos += strlen(start);

    Cor_CArg *nc_this = (Cor_CArg*) this;
    if (grp_arity())
        arg_ptr = nc_this->nested_args()->args();

    for (int i = grp_arity()-1; i >= 0; i--) {
        arg_ptr[i].sprint(str, pos, env);
        if (i > 0) {
            sprintf(str + *pos, ", ");
            *pos += 2;
        }
    }

    sprintf(str + *pos, end);
    *pos += strlen(end);
}

// *************************************************************************
void Cor_CArg::sprint(char *str, int *pos, Cor_BindEnv *context) const
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_CArg  *this2;

    switch (my_arg_kind) {

    default:
#ifdef    DEBUG
        TELL;
        cerr << "Unknown kind " << kind_str() << endl;
#endif
        CORAL_error(COR_FATAL_ERROR, "","sprint");
    case EMPTY_ARG:
#ifdef    DEBUG
        sprintf(str + *pos, "(Empty Arg)");
       *pos = strlen(str);
#endif
        break;

    case INT_ARG:
        sprintf(str + *pos, "%d", _int);
       *pos = strlen(str); 
        break;

    case DOUBLE_ARG:
        sprintf(str + *pos, "%f", _doub);
       *pos = strlen(str);
        break;

    case SYMBOL_ARG:
    case TINY_SYM_ARG:
        if (need_quotes(symbol()))
            sprintf(str + *pos, "\"%s\"", symbol());
        else
            sprintf(str + *pos, "%s", symbol());
       *pos = strlen(str);
        break;

    case CHUNK_ARG:
        sprintf(str + *pos, "%p", _chunk);
       *pos = strlen(str);
        break;

    case FUNCT_ARG:
        sprintf(str + *pos, "%s(", functor());
        *pos += symlen() + 1;
        this2 = (Cor_CArg *) this;
        this2->args().sprint(str, pos, context);
        sprintf(str + *pos, ")");
        (*pos)++;
        break;

    case VAR_ARG:
        if ((context == NULL) || context->lookup(_var_num).expr.is_empty()) {

            Cor_CArg *var_sym;
            if ((var_sym
                    = Cor_parserStruct->rule_var_names.get_ith(_var_num))) {
                sprintf(str + *pos, "%s", var_sym->symbol());
                *pos += var_sym->symlen();
            } else {
                sprintf(str + *pos, "X_%d_", _var_num);
                *pos += strlen(str + *pos);
            }

        } else {
            (context->lookup(_var_num).expr).sprint(str, pos,
                  context->lookup(_var_num).bindenv);
        }
        break;

    case REL_ARG:
        rel()->sprint(str,pos,context);
        break;

    case DONT_CARE_ARG:
        Cor_DontCareSymbol->sprint(str,pos,context);
        break;

    case PARAMETER_ARG:
        break;

      // Lists
    case LIST_ARG: {
        int j;
        sprintf(str + *pos, "[");
        (*pos)++;

        j = list._num_items - 1;
        ASSERT(j > 0);
        if (j >= 0) {
            Cor_CArg *car_arg;
            while (j > 1) {
                car_arg = list._items->get_ith(j--);
                car_arg->sprint(str,pos,context);
                sprintf(str + *pos, ",");
                (*pos)++;
            }
    
            Cor_CArg *last = list._items->get_ith(0);
            car_arg = list._items->get_ith(1);

            car_arg->sprint(str,pos,context);

            if (!last->is_nil()) {
                sprintf(str + *pos, "|");
                (*pos)++;
                last->sprint(str,pos,context);
            }
        }

        sprintf(str + *pos, "]");
        (*pos)++;
        break;
    }

    case NIL_ARG:
        sprintf(str + *pos, "[]");
        *pos += 2;
        break;

      // Arithmetic Expressions
    case ADD_OP_ARG:
    case SUB_OP_ARG:
    case MULT_OP_ARG:
    case DIV_OP_ARG:
    case IDIV_OP_ARG:
    case MOD_OP_ARG:

    case LSHFT_OP_ARG:
    case RSHFT_OP_ARG:

    case BT_AND_OP_ARG:
    case BT_OR_OP_ARG:
    case BT_NOT_OP_ARG:
    case BT_XOR_OP_ARG:
        sprintf(str + *pos, "<some number>");
        *pos += strlen("<some number>");
        break;

      // The various types of Grouping Args
    case SUM_GRPING_ARG:
        sprint_grping(str, pos, context, "sum(<", ">)");
        break;

    case PROD_GRPING_ARG:
        sprint_grping(str, pos, context, "prod(<", ">)");
        break;

    case CNT_GRPING_ARG:
        sprint_grping(str, pos, context, "count(<", ">)");
        break;

    case AVG_GRPING_ARG:
        sprint_grping(str, pos, context, "avg(<", ">)");
        break;

    case SUM_DGRPING_ARG:
        sprint_grping(str, pos, context, "sum_distinct(<", ">)");
        break;

    case PROD_DGRPING_ARG:
        sprint_grping(str, pos, context, "prod_distinct(<", ">)");
        break;

    case CNT_DGRPING_ARG:
        sprint_grping(str, pos, context, "count_distinct(<", ">)");
        break;

    case AVG_DGRPING_ARG:
        sprint_grping(str, pos, context, "avg_distinct(<", ">)");
        break;

    case MIN_GRPING_ARG:
        sprint_grping(str, pos, context, "min(<", ">)");
        break;

    case MAX_GRPING_ARG:
        sprint_grping(str, pos, context, "max(<", ">)");
        break;

    case SET_GRPING_ARG:
        sprint_grping(str, pos, context, "distinct(<", ">)");
        break;

    case ANY_GRPING_ARG:
        sprint_grping(str, pos, context, "any(<", ">)");
        break;

    case NO_AGG_GRPING_ARG:
        sprint_grping(str, pos, context, "<", ">");
        break;
    }
}

// *************************************************************************
void Cor_CArg::printon(ostream *file) const
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif

    ASSERT(this);

    char *fptr, format[20];
    char  buffer[100];

    switch (my_arg_kind) {


    default:
#ifdef    DEBUG
        TELL;
        cerr << "Unknown kind " << kind_str() << endl;
#endif
        CORAL_error(COR_FATAL_ERROR, "","printon1");
    case EMPTY_ARG:
#ifdef DEBUG
        cerr << "(Empty Arg)";
#endif
        return;

    case INT_ARG:
    case DOUBLE_ARG: 
        fptr = format;
  
        *fptr++ = '%';
        if (my_arg_kind == DOUBLE_ARG)
            *fptr++ = '#';
  
        if (print_width >= 0) {
            sprintf(fptr, "%ld", print_width);
            fptr += strlen(fptr);
        }
  
        if (print_precision >= 0) {
            sprintf(fptr, ".%ld", print_precision);
            fptr += strlen(fptr);
        }
  
        *fptr++ = ((my_arg_kind == INT_ARG) ?
              print_int_format : print_double_format);
        *fptr = '\0';
 
        if (my_arg_kind==INT_ARG)
            sprintf(buffer, format, _int);
        else
            sprintf(buffer, format, _doub);
   
          // Kludge to make sure the last char isn't '.'.
        fptr = buffer + strlen(buffer);
        if (fptr[-1] == '.') {
            fptr[0] = '0';
            fptr[1] = 0;
        }
  
        *file << buffer;
        return;

    case SYMBOL_ARG:
    case TINY_SYM_ARG:
        if (need_quotes(symbol())) {
            *file << "\"" << symbol() << "\"";
        } else {
            *file << symbol();
        }
        return;

    case CHUNK_ARG:
        *file << _chunk;
        return;

    case FUNCT_ARG:
    case VAR_ARG:
        break;

    case REL_ARG:
        rel()->print(0,file);
        return;

    case DONT_CARE_ARG:
        *file << *Cor_DontCareSymbol;
        return;

    case PARAMETER_ARG:
        *file << "$";
        return;

      // Lists
    case LIST_ARG:
    case NIL_ARG:
        break;

      // Arithmetic Expressions
    case ADD_OP_ARG:
    case SUB_OP_ARG:
    case MULT_OP_ARG:
    case DIV_OP_ARG:
    case IDIV_OP_ARG:
    case MOD_OP_ARG:

    case LSHFT_OP_ARG:
    case RSHFT_OP_ARG:

    case BT_AND_OP_ARG:
    case BT_OR_OP_ARG:
    case BT_NOT_OP_ARG:
    case BT_XOR_OP_ARG:
        break;

      // The various types of Grouping Args
    case SUM_GRPING_ARG:
    case PROD_GRPING_ARG:
    case CNT_GRPING_ARG:
    case AVG_GRPING_ARG:

    case SUM_DGRPING_ARG:
    case PROD_DGRPING_ARG:
    case CNT_DGRPING_ARG:
    case AVG_DGRPING_ARG:

    case MIN_GRPING_ARG:
    case MAX_GRPING_ARG:
    case SET_GRPING_ARG:
    case ANY_GRPING_ARG:
    case NO_AGG_GRPING_ARG:
        break;

    case AGG_RESULT_ARG:
        break;
    }

    print(0,file);
}

// *************************************************************************
// Print a char, converting chars to escape sequences.
static void print_metachr(char c, int in_str, ostream *file)
{
    switch (c) {
      case '\'':
        if (in_str) 
          *file << "'";
        else
          *file << "\\'";
        break;
      case '"':
        if (in_str) 
          *file << "\\\"";
        else
          *file << "\"";
        break;
      case '\0':
        *file << "\\0";
        break;
      case '\\':
        *file << "\\\\";
        break;
      case '\n':
        *file << "\\n";
        break;
      case '\t':
        *file << "\\t";
        break;
      case '\r':
        *file << "\\r";
        break;
      case '\f':
        *file << "\\f";
        break;
      case '\b':
        *file << "\\b";
        break;
      case '\v':
        *file << "\\v";
        break;
      case '\a':
        *file << "\\a";
        break;
      default:
        *file << c;
    }
}

// *************************************************************************
/* Print a string, converting chars to escape sequences. */
static void print_metastr(char *str, ostream *file)
{
  while (*str) {
    print_metachr(*str,1,file);
    str++;
  }
}

// *************************************************************************
void Cor_CArg::literal_print(ostream *file) const
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif

    if (is_symbol() && need_quotes(symbol())) {
      *file << "\"";
      print_metastr(symbol(),file);
      *file << "\"";
    } else
      printon(file); 
}

// *************************************************************************
// Stream printing
ostream& operator<< (ostream& out, const Cor_CArg& carg)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    switch (carg.my_arg_kind) {

    default:
#ifdef    DEBUG
        TELL;
        cerr << "Unknown kind " << carg.kind_str() << endl;
#endif
        CORAL_error(COR_FATAL_ERROR, "",
                __PRETTY_FUNCTION__, __FILE__, __LINE__);
        ASSERT(0);

    case EMPTY_ARG:
#ifdef    DEBUG
        cerr << "(Empty Arg)";
#endif
        break;

    case INT_ARG:
        out << carg._int;
        break;

    case DOUBLE_ARG:
        out << carg._doub;
        break;

    case SYMBOL_ARG:
    case TINY_SYM_ARG:
          // Need to check if they need quotes on that
        if (carg.need_quotes(carg.symbol())) {
            out << "\"";
            out << carg.symbol();
            out << "\"";
        } else 
            out << carg.symbol();
        break;

    case CHUNK_ARG:
        out << carg._chunk;
        break;

    case FUNCT_ARG:
    case VAR_ARG:
        carg.print(0,&out);
        break;

    case REL_ARG:
        carg._rel->print(0,&out);
        break;

    case DONT_CARE_ARG:
        out << *Cor_DontCareSymbol;
        break;

    case PARAMETER_ARG:
        out << "$";
        break;

      // Lists
    case LIST_ARG:
    case NIL_ARG:
        carg.print(0,&out);
        break;

      // Arithmetic Expressions
    case ADD_OP_ARG:
    case SUB_OP_ARG:
    case MULT_OP_ARG:
    case DIV_OP_ARG:
    case IDIV_OP_ARG:
    case MOD_OP_ARG:

    case LSHFT_OP_ARG:
    case RSHFT_OP_ARG:

    case BT_AND_OP_ARG:
    case BT_OR_OP_ARG:
    case BT_NOT_OP_ARG:
    case BT_XOR_OP_ARG:
        carg.print(0,&out);
        break;

      // The various types of Grouping Args
    case SUM_GRPING_ARG:
    case PROD_GRPING_ARG:
    case CNT_GRPING_ARG:
    case AVG_GRPING_ARG:

    case SUM_DGRPING_ARG:
    case PROD_DGRPING_ARG:
    case CNT_DGRPING_ARG:
    case AVG_DGRPING_ARG:

    case MIN_GRPING_ARG:
    case MAX_GRPING_ARG:
    case SET_GRPING_ARG:
    case ANY_GRPING_ARG:
    case NO_AGG_GRPING_ARG:
        carg.print(0,&out);
        break;

    case AGG_RESULT_ARG:
        carg.print(0,&out);
        break;
    }

    return(out);
}

// *************************************************************************
void Cor_CArg::print_binary_expr(Cor_BindEnv *context,
        ostream *file, char *op) const
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    *file << "(";
    expr._larg->print(context,file);
    *file << op;
    expr._rarg->print(context,file);
    *file << ")";
}

void Cor_CArg::print_expr_qv(Cor_BindEnv *context,
        ostream *file) const
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_CArg *me = ((Cor_CArg *) this)->dup();

    me->simplify_me(context);

    *file << *me;
}

void Cor_CArg::print_grping(Cor_BindEnv *context, ostream *file,
        char *start, char *end) const
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_ArgList  arg_ptr;

    *file << start;

    Cor_CArg *nc_this = (Cor_CArg*) this;
    if (grp_arity())
        arg_ptr = nc_this->nested_args()->args();

    for (int i=grp_arity()-1; i >= 0; i--) {
        arg_ptr[i].print(context, file);
        if (i > 0)
            *file << ", ";
    }

    *file << end;
}

// For Explain
void Cor_CArg::print_grping_qv(Cor_BindEnv *context, ostream *file,
        char *start, char *end) const
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_ArgList  arg_ptr;

    *file << start;

    Cor_CArg *nc_this = (Cor_CArg*) this;
    if (grp_arity())
        arg_ptr = nc_this->nested_args()->args();

    for (int i=grp_arity()-1; i >= 0; i--) {
        arg_ptr[i].print_qv(context, file);
        if (i > 0)
            *file << ", ";
    }

    *file << end;
}

// *************************************************************************
void Cor_CArg::print(Cor_BindEnv *context, ostream *file) const
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_CArg *this2;

    ASSERT(this);

    switch (my_arg_kind) {

    default:
#ifdef    DEBUG
        TELL;
        cerr << "Unknown kind " << kind_str() << endl;
        ASSERT(0);
#endif
    case EMPTY_ARG:
        printon(file);
        break;

    case INT_ARG:
    case DOUBLE_ARG:
    case SYMBOL_ARG:
    case TINY_SYM_ARG:
        printon(file);
        break;

    case CHUNK_ARG:
        printon(file);
        break;

    case FUNCT_ARG:
        *file << functor() << "(";
        this2 = (Cor_CArg *) this;
        this2->args().print(context, file);
        *file << ")";
        break;

    case VAR_ARG:

        if ((context == NULL) || context->lookup(_var_num).expr.is_empty()) {

            Cor_CArg *var_sym;
            if ((var_sym
                    = Cor_parserStruct->rule_var_names.get_ith(_var_num))) {
                (*file) << var_sym->symbol();
            } else {
                (*file) << "X_" << _var_num << "_";
            }
        } else
            context->lookup(_var_num).printon(file);
        return;

    case REL_ARG:
        if (_rel)
            _rel->print(context, file);
        return;

    case DONT_CARE_ARG:
        *file << *Cor_DontCareSymbol;
        return;

    case PARAMETER_ARG:
        *file << "$";
        return;

      // Lists
    case LIST_ARG: {
        int j;
        *file << "[";

        j = list._num_items - 1;
        if (j >= 0) {
            Cor_CArg *car_arg;
            while (j > 1) {
                car_arg = list._items->get_ith(j--);
                car_arg->print(context,file);
                *file << ",";
            }
    
            car_arg = list._items->get_ith(1);
            Cor_CArg *last = list._items->get_ith(0);

            if (car_arg)
                car_arg->print(context,file);
            else
                *file << "<null>";

            if (!last->is_nil()) {
                *file << "|";
                last->print(context,file);
            }
        }

        *file << "]";
        break;
    }

    case NIL_ARG:
        *file << "[]";
        break;

      // Arithmetic Expressions
    case ADD_OP_ARG:
        print_binary_expr(context, file, "+");
        break;

    case SUB_OP_ARG:
        // Need to check for unary minus
        if (expr._rarg)
            print_binary_expr(context, file, "-");
        else {
            *file << "(-";
            expr._larg->print(context,file);
            *file << ")";
        }
        break;

    case MULT_OP_ARG:
        print_binary_expr(context, file, "*");
        break;

    case DIV_OP_ARG:
        print_binary_expr(context, file, "/");
        break;

    case IDIV_OP_ARG:
        print_binary_expr(context, file, "//");
        break;

    case MOD_OP_ARG:
        print_binary_expr(context, file, " mod ");
        break;

    case LSHFT_OP_ARG:
        print_binary_expr(context, file, "<<");
        break;

    case RSHFT_OP_ARG:
        print_binary_expr(context, file, ">>");
        break;

    case BT_AND_OP_ARG:
        print_binary_expr(context, file, Cor_BitandOpSymbol->symbol());
        break;

    case BT_OR_OP_ARG:
        print_binary_expr(context, file, Cor_BitorOpSymbol->symbol());
        break;

    case BT_NOT_OP_ARG:
        print_binary_expr(context, file, Cor_BitnotOpSymbol->symbol());
        break;

    case BT_XOR_OP_ARG:
        print_binary_expr(context, file, Cor_BitxorOpSymbol->symbol());
        break;


      // The various types of Grouping Args
    case SUM_GRPING_ARG:
        print_grping(context, file, "sum(<", ">)");
        break;

    case PROD_GRPING_ARG:
        print_grping(context, file, "prod(<", ">)");
        break;

    case CNT_GRPING_ARG:
        print_grping(context, file, "count(<", ">)");
        break;

    case AVG_GRPING_ARG:
        print_grping(context, file, "avg(<", ">)");
        break;

    case SUM_DGRPING_ARG:
        print_grping(context, file, "sum_distinct(<", ">)");
        break;

    case PROD_DGRPING_ARG:
        print_grping(context, file, "prod_distinct(<", ">)");
        break;

    case CNT_DGRPING_ARG:
        print_grping(context, file, "count_distinct(<", ">)");
        break;

    case AVG_DGRPING_ARG:
        print_grping(context, file, "avg_distinct(<", ">)");
        break;

    case MIN_GRPING_ARG:
        print_grping(context, file, "min(<", ">)");
        break;

    case MAX_GRPING_ARG:
        print_grping(context, file, "max(<", ">)");
        break;

    case SET_GRPING_ARG:
        print_grping(context, file, "distinct(<", ">)");
        break;

    case ANY_GRPING_ARG:
        print_grping(context, file, "any(<", ">)");
        break;

    case NO_AGG_GRPING_ARG:
        print_grping(context, file, "<", ">");
        break;

    case AGG_RESULT_ARG:
        _agg_num->print(context,file);
        break;
    }
}

// quote variables (for explain)
void Cor_CArg::print_qv(Cor_BindEnv *context, ostream *file) const
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_CArg *this2;

    switch (my_arg_kind) {

    default:
#ifdef    DEBUG
        TELL;
        cerr << "Unknown kind " << kind_str() << endl;
        ASSERT(0);
#endif
    case EMPTY_ARG:
        printon(file);
        break;

    case INT_ARG:
    case DOUBLE_ARG:
    case SYMBOL_ARG:
    case TINY_SYM_ARG:
        printon(file);
        break;

    case CHUNK_ARG:
        printon(file);
        break;

    case FUNCT_ARG:
        *file << functor() << "(";
        this2 = (Cor_CArg *) this;
        this2->args().print_qv(context, file);
        *file << ")";
        break;

    case VAR_ARG:

        if ((context == NULL) || context->lookup(_var_num).expr.is_empty()) {

            Cor_CArg *var_sym;
            if ((var_sym
                    = Cor_parserStruct->rule_var_names.get_ith(_var_num))) {
                (*file) << "\"" << var_sym->symbol() << "\"";
            } else {
                (*file) << "\"X_" << _var_num << "_\"";
            }
        } else
            context->lookup(_var_num).printon(file);
        return;

    case REL_ARG:
        if (_rel)
            _rel->print(context, file);
        return;

    case DONT_CARE_ARG:
        *file << *Cor_DontCareSymbol;
        return;

    case PARAMETER_ARG:
        *file << "$";
        return;

      // Lists
    case LIST_ARG: {
        int j;
        *file << "[";

        j = list._num_items - 1;
        if (j >= 0) {
            Cor_CArg *car_arg;
            while (j > 1) {
                car_arg = list._items->get_ith(j--);
                car_arg->print_qv(context,file);
                *file << ",";
            }
    
            car_arg = list._items->get_ith(1);
            Cor_CArg *last = list._items->get_ith(0);

            car_arg->print_qv(context,file);

            if (!last->is_nil()) {
                *file << "|";
                last->print_qv(context,file);
            }
        }

        *file << "]";
        break;
    }

    case NIL_ARG:
        *file << "[]";
        break;

      // Arithmetic Expressions
      // Assume they are constant
    case ADD_OP_ARG:
    case SUB_OP_ARG:
    case MULT_OP_ARG:
    case DIV_OP_ARG:
    case IDIV_OP_ARG:
    case MOD_OP_ARG:
    case LSHFT_OP_ARG:
    case RSHFT_OP_ARG:
    case BT_AND_OP_ARG:
    case BT_OR_OP_ARG:
    case BT_NOT_OP_ARG:
    case BT_XOR_OP_ARG:
        print_expr_qv(context,file);
        break;


      // The various types of Grouping Args
    case SUM_GRPING_ARG:
        *file << "\"";
        print_grping_qv(context, file, "sum(<", ">)");
        *file << "\"";
        break;

    case PROD_GRPING_ARG:
        *file << "\"";
        print_grping_qv(context, file, "prod(<", ">)");
        *file << "\"";
        break;

    case CNT_GRPING_ARG:
        *file << "\"";
        print_grping_qv(context, file, "count(<", ">)");
        *file << "\"";
        break;

    case AVG_GRPING_ARG:
        *file << "\"";
        print_grping_qv(context, file, "avg(<", ">)");
        *file << "\"";
        break;

    case SUM_DGRPING_ARG:
        *file << "\"";
        print_grping_qv(context, file, "sum_distinct(<", ">)");
        *file << "\"";
        break;

    case PROD_DGRPING_ARG:
        *file << "\"";
        print_grping_qv(context, file, "prod_distinct(<", ">)");
        *file << "\"";
        break;

    case CNT_DGRPING_ARG:
        *file << "\"";
        print_grping_qv(context, file, "count_distinct(<", ">)");
        *file << "\"";
        break;

    case AVG_DGRPING_ARG:
        *file << "\"";
        print_grping_qv(context, file, "avg_distinct(<", ">)");
        *file << "\"";
        break;

    case MIN_GRPING_ARG:
        *file << "\"";
        print_grping_qv(context, file, "min(<", ">)");
        *file << "\"";
        break;

    case MAX_GRPING_ARG:
        *file << "\"";
        print_grping_qv(context, file, "max(<", ">)");
        *file << "\"";
        break;

    case SET_GRPING_ARG:
        *file << "\"";
        print_grping_qv(context, file, "distinct(<", ">)");
        *file << "\"";
        break;

    case ANY_GRPING_ARG:
        *file << "\"";
        print_grping_qv(context, file, "any(<", ">)");
        *file << "\"";
        break;

    case NO_AGG_GRPING_ARG:
        *file << "\"";
        print_grping_qv(context, file, "<", ">");
        *file << "\"";
        break;

    case AGG_RESULT_ARG:
        _agg_num->print(context,file);
        break;
    }
}

// *************************************************************************
void Cor_CArg::output() const
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    switch (my_arg_kind) {

    default:
#ifdef    DEBUG
        TELL;
        cerr << "Unknown kind " << kind_str() << endl;
#endif
    case EMPTY_ARG:
        break;

    case INT_ARG:
    case DOUBLE_ARG:
    case SYMBOL_ARG:
    case TINY_SYM_ARG:
        break;

    case CHUNK_ARG:
        break;

    case FUNCT_ARG:
        break;

    case VAR_ARG:
        break;

    case REL_ARG:
        if (_rel)
            _rel->print(0, NULL);
        return;

    case DONT_CARE_ARG:
    case PARAMETER_ARG:
        break;

      // Lists
    case LIST_ARG:
    case NIL_ARG:
        break;

      // Arithmetic Expressions
    case ADD_OP_ARG:
    case SUB_OP_ARG:
    case MULT_OP_ARG:
    case DIV_OP_ARG:
    case IDIV_OP_ARG:
    case MOD_OP_ARG:

    case LSHFT_OP_ARG:
    case RSHFT_OP_ARG:

    case BT_AND_OP_ARG:
    case BT_OR_OP_ARG:
    case BT_NOT_OP_ARG:
    case BT_XOR_OP_ARG:
        break;

      // The various types of Grouping Args
    case SUM_GRPING_ARG:
    case PROD_GRPING_ARG:
    case CNT_GRPING_ARG:
    case AVG_GRPING_ARG:

    case SUM_DGRPING_ARG:
    case PROD_DGRPING_ARG:
    case CNT_DGRPING_ARG:
    case AVG_DGRPING_ARG:

    case MIN_GRPING_ARG:
    case MAX_GRPING_ARG:
    case SET_GRPING_ARG:
    case ANY_GRPING_ARG:
    case NO_AGG_GRPING_ARG:
        break;

    case AGG_RESULT_ARG:
        break;
    }

    print(0, Cor_exEnv->out_strm);
}

// *************************************************************************
void Cor_CArg::print_dump(Cor_BindEnv *context, ostream *file) const
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    switch (my_arg_kind) {

    default:
#ifdef    DEBUG
        TELL;
        cerr << "Unknown kind " << kind_str() << endl;
#endif
    case EMPTY_ARG:
        break;

    case INT_ARG:
    case DOUBLE_ARG:
    case SYMBOL_ARG:
    case TINY_SYM_ARG:
        break;

    case CHUNK_ARG:
        break;

    case VAR_ARG:
    case FUNCT_ARG:
    case LIST_ARG:
        print_qv(context,file);
        return;

    case REL_ARG:
    case DONT_CARE_ARG:
    case PARAMETER_ARG:
        break;

      // Lists
    case NIL_ARG:
        break;

      // Arithmetic Expressions
    case ADD_OP_ARG:
    case SUB_OP_ARG:
    case MULT_OP_ARG:
    case DIV_OP_ARG:
    case IDIV_OP_ARG:
    case MOD_OP_ARG:

    case LSHFT_OP_ARG:
    case RSHFT_OP_ARG:

    case BT_AND_OP_ARG:
    case BT_OR_OP_ARG:
    case BT_NOT_OP_ARG:
    case BT_XOR_OP_ARG:
        break;

      // The various types of Grouping Args
    case SUM_GRPING_ARG:
    case PROD_GRPING_ARG:
    case CNT_GRPING_ARG:
    case AVG_GRPING_ARG:

    case SUM_DGRPING_ARG:
    case PROD_DGRPING_ARG:
    case CNT_DGRPING_ARG:
    case AVG_DGRPING_ARG:

    case MIN_GRPING_ARG:
    case MAX_GRPING_ARG:
    case SET_GRPING_ARG:
    case ANY_GRPING_ARG:
    case NO_AGG_GRPING_ARG:
        *file << "\"";
        print(context,file);
        *file << "\"";
        return;

    case AGG_RESULT_ARG:
        break;
    }

    print(context,file);
}

// *************************************************************************
void Cor_CArg::simplify_expr(Cor_CArg& carg,
            Cor_BindEnv  *context,
            Cor_TermLink *&renamed_vars,
            Cor_BindEnv  *const_env,
            Cor_BindEnv  *rename_only_env,
            ExprFunc     *efunc)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_CArg ltmp;
    Cor_CArg rtmp;

    larg()->simplify(ltmp, context,renamed_vars, const_env,
                rename_only_env);

    rarg()->simplify(rtmp, context, renamed_vars, const_env,
                rename_only_env);

    (efunc)(carg, ltmp, rtmp);
}

// *************************************************************************
void Cor_CArg::simplify(Cor_CArg& carg,
            Cor_BindEnv  *context,
            Cor_TermLink *&renamed_vars,
            Cor_BindEnv  *const_env,
            Cor_BindEnv  *rename_only_env)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif

    switch (my_arg_kind) {

    default:
#ifdef    DEBUG
        TELL;
        cerr << "Unknown kind " << kind_str() << endl;
#endif
    case EMPTY_ARG:
        break;

    case INT_ARG:
    case DOUBLE_ARG:
    case SYMBOL_ARG:
    case TINY_SYM_ARG:
        break;

    case CHUNK_ARG:
        break;

    case FUNCT_ARG: {

            int changes = 0;
            if (isConstant()) {
                carg = *this;
                return;
            }

            if (rename_only_env) {
                if ((context != const_env) && (context != rename_only_env)){
                    carg.init();
                    return;
                }

           // WARNING:
           //  This assumes that if rename_only_var is non-null, the function
           //  has been called recursively from backpatch_trail.
           //  In this case, there is no need to follow terms that are
           //  interpreted in const_env.

                if (context == const_env) {
                    carg = *this;
                    return;
                }
            }

            Cor_CArg tmp(functor());
            Cor_CArg *this_copy = new Cor_CArg(tmp, arity());

            int i= 0;
            FOR_EACH_ARG(arg, args()) {
                         arg->simplify(this_copy->args()[i], context,
                         renamed_vars, const_env, rename_only_env);
                if ((this_copy->args()[i]).is_empty()) {
                    delete this_copy;
                    carg.init();
                    return;
                }

                if (!(this_copy->args()[i]).equals(*arg))
                    changes = 1;
                i++;

            } END_EACH_ARG;

            if (changes) {
                // computes, and if the term is ground,
                // stores the hash value
                this_copy->_hash = UnknownHashValue;   // SPF-ADD
                this_copy->hash();
                carg = *this_copy;
            } else {
                carg = *this;
            }

            delete this_copy;
        }
        return;

    case VAR_ARG: {

        if (rename_only_env && (context != rename_only_env)
                && (context != const_env)) {
           carg.init();
           return;
        }
   
        Cor_Term term;
        if (context != NULL) {
            term = context->lookup(_var_num);
        }
   
        if ((context == const_env)
                && ((context == NULL) || term.expr.is_empty()) ) {
            carg = *this;
            return;
        }

        if ((context == NULL) || term.expr.is_empty()) {

            Cor_TermLink *tlink = renamed_vars;
            int new_i = 0;
            for ( ; tlink != NULL; tlink = tlink->next, new_i++)
                if ((tlink->old_var == _var_num)
                            && (tlink->old_bindenv == context)) {
                    carg = tlink->new_arg;
                    return;
                }

            tlink = new Cor_TermLink();
            tlink->next = renamed_vars;
            renamed_vars = tlink;
            tlink->old_var = _var_num;
            tlink->old_bindenv = context;
            tlink->new_arg = Cor_CArg();
            tlink->new_arg.set_var(const_env->max_var_number()+new_i);
            carg = tlink->new_arg;
            return;
        }

        term.expr.simplify( carg, term.bindenv, renamed_vars, const_env,
                   rename_only_env);
        }
        return;

    case REL_ARG:
    case DONT_CARE_ARG:
        break;

    case PARAMETER_ARG:
        TELL;
        cerr << "Need something here yet?\n";
        break;

      // Lists
    case LIST_ARG: {

            if (Cor_isGroundHashValue(hash())) {
                break;
            }

            Cor_CArg *cpy = this->list_dup();

            int j;
            for (j = cpy->list._num_items-1; j >= 0; j--) {
                Cor_CArg *tmp = cpy->list._items->get_ith(j);
                tmp->simplify_me(context);
            }

            cpy =  cpy->normalize_list();

            carg.my_arg_kind = cpy->my_arg_kind;
            carg.list._items = cpy->list._items;
            carg.list._num_items = cpy->list._num_items;
            carg._hash = UnknownHashValue; // Probably messed up

            cpy->init();
            delete cpy;
        }
        return;

    case NIL_ARG:
        break;

      // Arithmetic Expressions
    case ADD_OP_ARG:
        simplify_expr(carg, context, renamed_vars, const_env,
                rename_only_env, Cor_AddOp);
        return;

    case SUB_OP_ARG:
        if (! rarg()) {
            Cor_CArg ltmp;
            larg()->simplify(ltmp,context,renamed_vars, const_env,
                    rename_only_env);
            Cor_SubOp(carg, Cor_ZeroCArg, ltmp);
        } else
            simplify_expr(carg, context, renamed_vars, const_env,
                    rename_only_env, Cor_SubOp);
    
        return;

    case MULT_OP_ARG:
        simplify_expr(carg, context, renamed_vars, const_env,
                rename_only_env, Cor_MulOp);
        return;

    case DIV_OP_ARG:
        simplify_expr(carg, context, renamed_vars, const_env,
                rename_only_env, Cor_DivOp);
        return;

    case IDIV_OP_ARG:
        simplify_expr(carg, context, renamed_vars, const_env,
                rename_only_env, Cor_DivOp);
        return;

    case MOD_OP_ARG:
        simplify_expr(carg, context, renamed_vars, const_env,
                rename_only_env, Cor_ModOp);
        return;

    case LSHFT_OP_ARG:
        simplify_expr(carg, context, renamed_vars, const_env,
                rename_only_env, Cor_LShiftOp);
        return;

    case RSHFT_OP_ARG:
        simplify_expr(carg, context, renamed_vars, const_env,
                rename_only_env, Cor_RShiftOp);
        return;

    case BT_AND_OP_ARG:
        simplify_expr(carg, context, renamed_vars, const_env,
                rename_only_env, Cor_BitandOp);
        return;

    case BT_OR_OP_ARG:
        simplify_expr(carg, context, renamed_vars, const_env,
                rename_only_env, Cor_BitorOp);
        return;

    case BT_NOT_OP_ARG: {
        Cor_CArg ltmp;
        larg()->simplify(ltmp,context,renamed_vars, const_env,
                rename_only_env);
        Cor_BitxorOp(carg, Cor_MinusOneCArg, ltmp);
        }
        return;

    case BT_XOR_OP_ARG:
        simplify_expr(carg, context, renamed_vars, const_env,
                rename_only_env, Cor_BitxorOp);
        return;


      // The various types of Grouping Args

    case SET_GRPING_ARG:
    case NO_AGG_GRPING_ARG: {

        Cor_StorageRelation *rel =
            new Cor_HashSimpleRelation(NESTED_SET_ARITY, 0,
                       COR_DEFAULT_SET_BUCKETS);
        rel->name = *Cor_NestedSymbol;

        Cor_ArgList *copy_args = new Cor_ArgList(grp_arity());
        // WARNING - the following may not be quite right
        if (grp_arity())
            Cor_CopyArgs(copy_args, &(nested_args()->args()),
                    context, (int) grp_arity(), const_env, rename_only_env);
        else
            Cor_CopyArgs(copy_args, NULL, context, (int) grp_arity(),
                    const_env, rename_only_env);

        Cor_Tuple *tuple1 = new Cor_Tuple(copy_args);

        int inserted = 0;

        if (Kind() == SET_GRPING_ARG) {
            if (!rel->is_subsumed(tuple1))
                inserted = rel->insert_tuple(tuple1);
        } else {
            inserted = rel->insert_tuple(tuple1);
        }

        if (!inserted) {
            delete tuple1;
            delete copy_args;
        }

        carg = Cor_CArg(rel);
        return;
        }

    case CNT_GRPING_ARG:
        carg = Cor_OneCArg;
        return;

    case SUM_GRPING_ARG:
    case PROD_GRPING_ARG:
    case MIN_GRPING_ARG:
    case MAX_GRPING_ARG:
    case ANY_GRPING_ARG: {

        if (grp_arity() != 1) {
            *(Cor_exEnv->err_strm) << "Grouping::simplify: bad arity\n";
            carg = Cor_ZeroCArg;
            return;
        }

        (nested_args()->args())[0].simplify(carg, context,
                    renamed_vars, const_env, rename_only_env);

        return;
        }


    case AVG_GRPING_ARG: {

        if (grp_arity() != 1) {
            *(Cor_exEnv->err_strm) << "Grouping::simplify: bad arity\n";
            carg = Cor_ZeroCArg;
            return;
        }

        Cor_CArg new_val;
        (nested_args()->args())[0].simplify(new_val, context,
                    renamed_vars, const_env, rename_only_env);

        if (!new_val.is_num()) {

            *(Cor_exEnv->err_strm)
                    << "Avg aggregation not a numeric constant!\n";
            carg = 0;
            return;
        }

        carg.set_kind(AGG_RESULT_ARG);
        carg._agg_num = new Cor_CArg(new_val.NumDval());
        carg.agg._agg_count = 1;
        carg.agg._dist_rel  = NULL;
        }
        return;

    case CNT_DGRPING_ARG: {

        if (grp_arity() != 1) {
            *(Cor_exEnv->err_strm) << "Grouping::simplify: bad arity\n";
            carg = 0;
            return;
        }

        Cor_CArg new_valc;
        (nested_args()->args())[0].simplify(new_valc, context,
                       renamed_vars, const_env, rename_only_env);
        Cor_StorageRelation *new_relc =
            new Cor_HashSimpleRelation(NESTED_SET_ARITY, 0,
                           COR_DEFAULT_SET_BUCKETS);
        new_relc->name = *Cor_NestedSymbol;
        new_relc->set_subsumption(1);
        Cor_ArgList *copy_argsc = new Cor_ArgList(1);
        (*copy_argsc)[0] = new_valc;
        Cor_Tuple *new_tuplec = new Cor_Tuple(copy_argsc);
        new_relc->insert_tuple(new_tuplec);

        carg.set_kind(AGG_RESULT_ARG);
        carg._agg_num = new Cor_CArg(1);
        carg.agg._agg_count = 0;    // Unused
        carg.agg._dist_rel = new_relc;
        (carg.agg._dist_rel->carg_count)++;
        }
        return;

    case AVG_DGRPING_ARG: {

        if (grp_arity() != 1) {
            *(Cor_exEnv->err_strm) << "Grouping::simplify: bad arity\n";
            carg = 0;
            return;
        }

        Cor_CArg new_vala;
        (nested_args()->args())[0].simplify(new_vala, context,
                    renamed_vars, const_env, rename_only_env);

        if (!new_vala.is_num()) {

            *(Cor_exEnv->err_strm) << "Aggregation on non-numeric constant!\n";
            carg = 0;
            return;
        }

        Cor_StorageRelation *new_rela =
            new Cor_HashSimpleRelation(NESTED_SET_ARITY, 0,
                       COR_DEFAULT_SET_BUCKETS);
        new_rela->name = *Cor_NestedSymbol;
        new_rela->set_subsumption(1);
        Cor_ArgList *copy_argsa = new Cor_ArgList(1);
        (*copy_argsa)[0] = new_vala;
        Cor_Tuple *new_tuplea = new Cor_Tuple(copy_argsa);
        new_rela->insert_tuple(new_tuplea);

        carg.set_kind(AGG_RESULT_ARG);
        carg._agg_num = new Cor_CArg(new_vala.NumDval());
        carg.agg._agg_count = 1;
        carg.agg._dist_rel = new_rela;
        (carg.agg._dist_rel->carg_count)++;
        }
        return;

    case PROD_DGRPING_ARG:
    case SUM_DGRPING_ARG: {

        if (grp_arity() != 1) {
            *(Cor_exEnv->err_strm) << "Grouping::simplify: bad arity\n";
            carg = 0;
            return;
        }

        Cor_CArg new_val2;
        (nested_args()->args())[0].simplify(new_val2, context,
                    renamed_vars, const_env, rename_only_env);

        if (!new_val2.is_num()) {
            *(Cor_exEnv->err_strm) << "Aggregation not a numeric constant!\n";
            carg = 0;
            return;
        }

        Cor_StorageRelation *new_rel
                = new Cor_HashSimpleRelation(NESTED_SET_ARITY, 0,
                         COR_DEFAULT_SET_BUCKETS);
        new_rel->name = *Cor_NestedSymbol;
        new_rel->set_subsumption(1);
        Cor_ArgList *copy_args = new Cor_ArgList(1);
        (*copy_args)[0] = new_val2;
        Cor_Tuple *new_tuple = new Cor_Tuple(copy_args);
        new_rel->insert_tuple(new_tuple);

        carg.set_kind(AGG_RESULT_ARG);
        carg._agg_num = new Cor_CArg(new_val2.NumDval());
        carg.agg._dist_rel = new_rel;
        (carg.agg._dist_rel->carg_count)++;
        }
        return;

    case AGG_RESULT_ARG:
        carg.set_kind(AGG_RESULT_ARG);
        carg._agg_num = _agg_num->dup();
        carg.agg._agg_count = agg._agg_count;
        carg.agg._dist_rel  = agg._dist_rel;
        if (carg.agg._dist_rel)
            (carg.agg._dist_rel->carg_count)++;
        break;
    }

    carg = *this;
}

// *************************************************************************
// Special case for simplifying yourself, return value is 0
// if variables still remain, 1 if new value is a constant.
int Cor_CArg::simplify_me(Cor_BindEnv *context)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_CArg *me;

    switch (my_arg_kind) {

    default:
#ifdef    DEBUG
        TELL;
        cerr << "Unknown kind " << kind_str() << endl;
#endif
    case EMPTY_ARG:
        break;

    case INT_ARG:
    case DOUBLE_ARG:
        break;

    case SYMBOL_ARG:
    case TINY_SYM_ARG:
        ASSERT(_hash != UnknownHashValue);
        break;

    case CHUNK_ARG:
        break;

    case FUNCT_ARG: {

            if (isConstant())
                return(1);

            int now_const = 1;
            FOR_EACH_ARG(arg, args()) {
                now_const &= arg->simplify_me(context);
            } END_EACH_ARG;

                // if the term is ground, compute and 
                // store the hash value
            if (now_const)
              {
              _hash = UnknownHashValue;   // SPF-ADD
              hash();
              }

            return(now_const); 
        }

    case VAR_ARG: {

            me = (Cor_CArg *) this;

            if (context) {
                Cor_Term term;
                if (context->try_lookup(term,_var_num)) {

                    *me = term.expr;

                    switch (me->Kind()) {

                    case SYMBOL_ARG:
                          // Symbols are too stupid to hash themselves
                        _hash = Cor_StrHash(sym._symbol);
                        break;
                    case TINY_SYM_ARG:
                          // Symbols are too stupid to hash themselves
                        _hash = Cor_StrHash(_symbuf);
                        break;

                    case VAR_ARG:
                    case SUM_GRPING_ARG:
                    case PROD_GRPING_ARG:
                    case CNT_GRPING_ARG: 
                    case AVG_GRPING_ARG:  
                    case STDDEV_GRPING_ARG:
                    case SUM_DGRPING_ARG:
                    case PROD_DGRPING_ARG:
                    case CNT_DGRPING_ARG:
                    case AVG_DGRPING_ARG:
                    case MIN_GRPING_ARG: 
                    case MAX_GRPING_ARG:
                    case SET_GRPING_ARG:
                    case ANY_GRPING_ARG:
                    case NO_AGG_GRPING_ARG:
                    case USR_DEFD_GRPING_ARG:

                    case AGG_RESULT_ARG:
                    case CMPLX_AGG_RESULT_ARG:

                    case DONT_CARE_ARG:

                          // These things don't use _hash
                        break;

                    default:
                        _hash = UnknownHashValue;
                        break;
                    }

                    ASSERT(me->_var_num >= 0);
                    return(me->simplify_me(term.bindenv));
                }
            }

            return(0);
        }

    case REL_ARG:
    case DONT_CARE_ARG:
        break;

    case PARAMETER_ARG:
        TELL;
        cerr << "Need something here yet?\n";
        break;

      // Lists
    case LIST_ARG: {

            if (Cor_isGroundHashValue(hash()))
                return(1);

            int now_const = 1;
            int j;
            for (j = list._num_items-1; j >= 0; j--) {
                Cor_CArg *tmp = list._items->get_ith(j);
                now_const &= tmp->simplify_me(context);
            }

            // Need to normalize somehow (using this)
            //cpy =  cpy->normalize_list();

            if (now_const)
                _hash = hash();

            return(now_const);
        }

    case NIL_ARG:
        break;

      // Arithmetic Expressions
    case ADD_OP_ARG:
        return(simplify_me_expr(context, Cor_AddOp));

    case SUB_OP_ARG: {
            me = (Cor_CArg *) this;
            int now_const = 1;
            if (! rarg()) {
                now_const &= larg()->simplify_me(context);
                Cor_SubOp(*me, Cor_ZeroCArg, *(larg()));
                return(now_const);
            } else
                return(simplify_me_expr(context, Cor_SubOp));
        }

    case MULT_OP_ARG:
        return(simplify_me_expr(context, Cor_MulOp));

    case DIV_OP_ARG:
        return(simplify_me_expr(context, Cor_DivOp));

    case IDIV_OP_ARG:
        return(simplify_me_expr(context, Cor_DivOp));

    case MOD_OP_ARG:
        return(simplify_me_expr(context, Cor_ModOp));

    case LSHFT_OP_ARG:
        return(simplify_me_expr(context, Cor_LShiftOp));

    case RSHFT_OP_ARG:
        return(simplify_me_expr(context, Cor_RShiftOp));

    case BT_AND_OP_ARG:
        return(simplify_me_expr(context, Cor_BitandOp));

    case BT_OR_OP_ARG:
        return(simplify_me_expr(context, Cor_BitorOp));

    case BT_NOT_OP_ARG: {
            me = (Cor_CArg *) this;
            int now_const = 1;
            now_const &= larg()->simplify_me(context);
            Cor_BitxorOp(*me, Cor_MinusOneCArg, *(larg()));
            return(now_const);
        }

    case BT_XOR_OP_ARG:
        return(simplify_me_expr(context, Cor_BitxorOp));


      // The various types of Grouping Args

    case SET_GRPING_ARG:
    case NO_AGG_GRPING_ARG: {

#if    0
        int env_size = 0;

        Cor_StorageRelation *rel =
            new Cor_HashSimpleRelation(NESTED_SET_ARITY, 0,
                       COR_DEFAULT_SET_BUCKETS);
        rel->name = *Cor_NestedSymbol;

        Cor_ArgList *copy_args = new Cor_ArgList(grp_arity());
        // WARNING - the following may not be quite right
        if (grp_arity())
            env_size = Cor_CopyArgs(copy_args, &(nested_args()->args()),
                        context, (int) grp_arity(), const_env,
                        rename_only_env);
        else
            env_size = Cor_CopyArgs(copy_args, NULL,
                        context, (int) grp_arity(), const_env,
                        rename_only_env);

        Cor_Tuple *tuple1 = new Cor_Tuple(copy_args);

        int inserted = 0;

        if (Kind() == SET_GRPING_ARG) {
            if (!rel->is_subsumed(tuple1))
                inserted = rel->insert_tuple(tuple1);
        } else {
            inserted = rel->insert_tuple(tuple1);
        }

        if (!inserted) {
            delete tuple1;
            delete copy_args;
        }

        carg = Cor_CArg(rel);
        return;
#endif
        TELL;
        cerr << "Need more code here.\n";
        return(1);
        }

    case CNT_GRPING_ARG: {
            me = (Cor_CArg *) this;
            *me = Cor_OneCArg;
            return(1);
        }
    case SUM_GRPING_ARG:
    case PROD_GRPING_ARG:
    case MIN_GRPING_ARG:
    case MAX_GRPING_ARG:
    case ANY_GRPING_ARG:

        TELL;
        cerr << "Need more code here.\n";
        return(1);
#if    0
        if (grp_arity() != 1) {
            *(Cor_exEnv->err_strm) << "Grouping::simplify: bad arity\n";
            carg = Cor_ZeroCArg;
            return;
        }

        (nested_args()->args())[0].simplify(carg, context,
                    renamed_vars, const_env, rename_only_env);
        return;
#endif

    case AVG_GRPING_ARG: {

        TELL;
        cerr << "Need more code here.\n";
        return(1);
#if    0
        if (grp_arity() != 1) {
            *(Cor_exEnv->err_strm) << "Grouping::simplify: bad arity\n";
            carg = Cor_ZeroCArg;
            return;
        }

        Cor_CArg new_val;
        (nested_args()->args())[0].simplify(new_val, context,
                    renamed_vars, const_env, rename_only_env);

        if (!new_val.is_num()) {

            *(Cor_exEnv->err_strm)
                    << "Avg aggregation not a numeric constant!\n";
            carg = 0;
            return;
        }

        carg.set_kind(AGG_RESULT_ARG);
        carg._agg_num = new Cor_CArg(new_val.NumDval());
        carg.agg._agg_count = 1;
        carg.agg._dist_rel  = NULL;
#endif
        }

    case CNT_DGRPING_ARG: {

        TELL;
        cerr << "Need more code here.\n";
        return(1);
#if    0
        if (grp_arity() != 1) {
            *(Cor_exEnv->err_strm) << "Grouping::simplify: bad arity\n";
            carg = 0;
            return;
        }

        Cor_CArg new_valc;
        (nested_args()->args())[0].simplify(new_valc, context,
                       renamed_vars, const_env, rename_only_env);
        Cor_StorageRelation *new_relc =
            new Cor_HashSimpleRelation(NESTED_SET_ARITY, 0,
                           COR_DEFAULT_SET_BUCKETS);
        new_relc->name = *Cor_NestedSymbol;
        new_relc->set_subsumption(1);
        Cor_ArgList *copy_argsc = new Cor_ArgList(1);
        (*copy_argsc)[0] = new_valc;
        Cor_Tuple *new_tuplec = new Cor_Tuple(copy_argsc, NULL);
        new_relc->insert_tuple(new_tuplec);

        carg.set_kind(AGG_RESULT_ARG);
        carg._agg_num = new Cor_CArg(1);
        carg.agg._agg_count = 0;    // Unused
        carg.agg._dist_rel = new_relc;
        (carg.agg._dist_rel->carg_count)++;
#endif
        }

    case AVG_DGRPING_ARG: {

        TELL;
        cerr << "Need more code here.\n";
        return(1);
#if    0
        if (grp_arity() != 1) {
            *(Cor_exEnv->err_strm) << "Grouping::simplify: bad arity\n";
            carg = 0;
            return;
        }

        Cor_CArg new_vala;
        (nested_args()->args())[0].simplify(new_vala, context,
                    renamed_vars, const_env, rename_only_env);

        if (!new_vala.is_num()) {

            *(Cor_exEnv->err_strm)
                << "Aggregation on non-numeric constant!\n";
            carg = 0;
            return;
        }

        Cor_StorageRelation *new_rela =
            new Cor_HashSimpleRelation(NESTED_SET_ARITY, 0,
                       COR_DEFAULT_SET_BUCKETS);
        new_rela->name = *Cor_NestedSymbol;
        new_rela->set_subsumption(1);
        Cor_ArgList *copy_argsa = new Cor_ArgList(1);
        (*copy_argsa)[0] = new_vala;
        Cor_Tuple *new_tuplea = new Cor_Tuple(copy_argsa, NULL);
        new_rela->insert_tuple(new_tuplea);

        carg.set_kind(AGG_RESULT_ARG);
        carg._agg_num = new Cor_CArg(new_vala.NumDval());
        carg.agg._agg_count = 1;
        carg.agg._dist_rel = new_rela;
        (carg.agg._dist_rel->carg_count)++;
#endif
        }

    case PROD_DGRPING_ARG:
    case SUM_DGRPING_ARG: {

        TELL;
        cerr << "Need more code here.\n";
        return(1);
#if    0
        if (grp_arity() != 1) {
            *(Cor_exEnv->err_strm) << "Grouping::simplify: bad arity\n";
            carg = 0;
            return;
        }

        Cor_CArg new_val2;
        (nested_args()->args())[0].simplify(new_val2, context,
                    renamed_vars, const_env, rename_only_env);

        if (!new_val2.is_num()) {
            *(Cor_exEnv->err_strm) << "Aggregation not a numeric constant!\n";
            carg = 0;
            return;
        }

        Cor_StorageRelation *new_rel
                = new Cor_HashSimpleRelation(NESTED_SET_ARITY, 0,
                         COR_DEFAULT_SET_BUCKETS);
        new_rel->name = *Cor_NestedSymbol;
        new_rel->set_subsumption(1);
        Cor_ArgList *copy_args = new Cor_ArgList(1);
        (*copy_args)[0] = new_val2;
        Cor_Tuple *new_tuple = new Cor_Tuple(copy_args, NULL);
        new_rel->insert_tuple(new_tuple);

        carg.set_kind(AGG_RESULT_ARG);
        carg._agg_num = new Cor_CArg(new_val2.NumDval());
        carg.agg._dist_rel = new_rel;
        (carg.agg._dist_rel->carg_count)++;
#endif
        }

    case AGG_RESULT_ARG:
        TELL;
        cerr << "Need more code here.\n";
        return(1);
#if    0
        carg.set_kind(AGG_RESULT_ARG);
        carg._agg_num = _agg_num->dup();
        carg.agg._agg_count = agg._agg_count;
        carg.agg._dist_rel  = agg._dist_rel;
        if (carg.agg._dist_rel)
            (carg.agg._dist_rel->carg_count)++;
        break;
#endif
    }

    return(1);
}

// *************************************************************************
// Special case for simplifying yourself, return value is 0
// if variables still remain, 1 if new value is a constant.
int Cor_CArg::simplify_me_expr(Cor_BindEnv *context, ExprFunc *efunc)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int now_const = 1;
    Cor_CArg *me = (Cor_CArg *) this;

    now_const &= larg()->simplify_me(context);
    now_const &= rarg()->simplify_me(context);

    (efunc)(*me, *(larg()), *(rarg()));

    return(now_const);
}

// *************************************************************************
// Update the grouping argument of the tuple in place.
// For reasons beyond my understanding (but within Raghu's), this will
// never be a problem.  In his defense, and to my surprise, it works.
//
int Cor_CArg::update_tuple(Cor_Tuple *old_tuple, Cor_CArg val,
                int argnum, Cor_BindEnv *context)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    switch (Kind()) {

    default:
#ifdef    DEBUG
        TELL;
        cerr << "Unknown kind for grouping." << kind_str() << endl;
#endif
        ASSERT(0);
        return(0);

    case SET_GRPING_ARG:
    case NO_AGG_GRPING_ARG:
    {
        if (! old_tuple->args()[argnum].is_rel()) {
          *(Cor_exEnv->err_strm)
             << "Error: grouping - expected relation argument to insert into\n";
            return(0);
        }

        Cor_StorageRelation *rel
            = (Cor_StorageRelation *) old_tuple->args()[argnum].rel();

        rel->_hash = UnknownHashValue; 
                // so that a new hash value will be computed for the set
        Cor_ArgList *copy_args = new Cor_ArgList(grp_arity());

        /*********************************
        // WARNING::  ReturnUnify optimizations needed here!!!
        ....
        .... See how this is affected by and will affect return unify 
        .... optimizations
        ....
        **********************************/

        Cor_CopyArgs(copy_args, &(nested_args()->args()),
            context, grp_arity(), Cor_UnusedEnv);
        Cor_Tuple *tuple1 = new Cor_Tuple(copy_args);

        if (Kind() == SET_GRPING_ARG) {
            if (!rel->is_subsumed(tuple1))
                rel->insert_tuple(tuple1);
        } else
            rel->insert_tuple(tuple1);

        // NOTE: add rel->env_size and update it here.
        // also add rel->hash_val and rel->id and update them
        // also do this whenever nested relations are created.
        // (see Gram.y)

        return(0);
    }

    case CNT_GRPING_ARG: {
        if (grp_arity() != 1) {
            *(Cor_exEnv->err_strm)
                << "Error: count(<>) aggregation: wrong arity.\n";
            return(0);
        }
     
        int itmp = old_tuple->args()[argnum].NumIval() + 1;
        old_tuple->args()[argnum] = itmp;

        return(1);
    }

    case ANY_GRPING_ARG: {

        if (grp_arity() != 1) {
            *(Cor_exEnv->err_strm)
                << "Error: any(<>) aggregation: wrong arity.\n";
            return(0);
        }
        return(1);
    }

    case SUM_GRPING_ARG: {

        if (grp_arity() != 1) {
            *(Cor_exEnv->err_strm)
                << "Error: sum(<>) aggregation: wrong arity.\n";
            return(0);
        }

        if (! old_tuple->args()[argnum].is_num()
                    || ! val.is_num() ) {
            *(Cor_exEnv->err_strm)
                << "Error: sum(<>) aggregation: not a number.\n";
            return(0);
        }

        double dtmp = old_tuple->args()[argnum].NumDval() + val.NumDval();
        old_tuple->args()[argnum] = dtmp;

        return(1);
    }

    case PROD_GRPING_ARG: {

        if (grp_arity() != 1) {
            *(Cor_exEnv->err_strm)
                << "Error: prod(<>) aggregation: wrong arity.\n";
            return(0);
        }

        if (! old_tuple->args()[argnum].is_num()
                    || ! val.is_num() ) {

            *(Cor_exEnv->err_strm)
                << "Error: prod(<>) aggregation: not a number.\n";
               return(0);
        }
                

        double dtmp = old_tuple->args()[argnum].NumDval() * val.NumDval();
        old_tuple->args()[argnum] = dtmp;

        return(1);
    }

    case MIN_GRPING_ARG:
    case MAX_GRPING_ARG:
    {
        if (grp_arity() != 1) {
            *(Cor_exEnv->err_strm)
                << "Error: min(<>) or max(<>) aggregation: wrong arity.\n";
            return(0);
        }

        if (! old_tuple->args()[argnum].is_num() ) {
            *(Cor_exEnv->err_strm)
                << "Error: min(<>) or max(<>) aggregation: not a number.\n";
            return(0);
        }

        int c_stat = old_tuple->args()[argnum].compare_num(val);

        if (Kind() == MAX_GRPING_ARG)
            c_stat = -c_stat;

        if (c_stat <= 0) {         // No replacement to be done.
            return(0);
        }

#ifdef DO_TRACE
        if (Cor_exEnv->dbg_aggregates)
            *(Cor_exEnv->trace_strm)
                << "Min/Max aggregation - new extreme value found\n";
#endif

        if ( val.is_int() )
            old_tuple->args()[argnum] = val.NumIval();
        else
            old_tuple->args()[argnum] = val.NumDval();

        return(1);
    }

    case AVG_GRPING_ARG: {

        if (grp_arity() != 1) {
            *(Cor_exEnv->err_strm)
                << "Error: avg(<>) aggregation: wrong arity.\n";
            return(0);
        }

        if (! old_tuple->args()[argnum]._agg_num->is_num()
                    || !val.is_num() ) {
            *(Cor_exEnv->err_strm)
                << "Error: avg(<>) aggregation: not a number.\n";
            return(0);
        }
                
        int cnt = old_tuple->args()[argnum].agg._agg_count;
        int new_cnt = cnt + 1;

        double av =  ( (double) cnt
            * old_tuple->args()[argnum]._agg_num->NumDval()
            + val.NumDval() ) / (double) new_cnt;

        old_tuple->args()[argnum].set_kind(AGG_RESULT_ARG);
        *(old_tuple->args()[argnum]._agg_num) = av;
        old_tuple->args()[argnum].agg._agg_count = new_cnt;
        old_tuple->args()[argnum].agg._dist_rel  = NULL;
        
        return(1);
    }

    case AVG_DGRPING_ARG:
    {
        if (grp_arity() != 1) {
            *(Cor_exEnv->err_strm)
                << "Error: avg_distinct(<>) aggregation: wrong arity.\n";
            return(0);
        }

        if (!old_tuple->args()[argnum]._agg_num->is_num()
                    || !val.is_num() ) {
            *(Cor_exEnv->err_strm)
                << "Error: avg_distinct(<>) aggregation: not a number.\n";
            return(0);
        }
                
        Cor_ArgList *val_lista = new Cor_ArgList(1);
        (*val_lista)[0] = val;
        Cor_Tuple *val_tuplea = new Cor_Tuple(val_lista);

        Cor_StorageRelation *drela = old_tuple->args()[argnum].dist_rel();

        if (drela->is_subsumed(val_tuplea)) {
           // then not a distinct element, so don't need to do anything.
           return(1);
        }

        // otherwise, need to update the grouping value, and the set

        drela->insert_tuple(val_tuplea);
        
        int cnt = old_tuple->args()[argnum].agg._agg_count;
        int new_cnt = cnt + 1;

        double av =  ( (double) cnt
            * old_tuple->args()[argnum]._agg_num->NumDval()
            + val.NumDval() ) / (double) new_cnt;

        old_tuple->args()[argnum].set_kind(AGG_RESULT_ARG);
        *(old_tuple->args()[argnum]._agg_num) = av;
        old_tuple->args()[argnum].agg._agg_count = new_cnt;
        old_tuple->args()[argnum].agg._dist_rel  = drela;

        return(1);
    }

    case CNT_DGRPING_ARG:
    {
        if (grp_arity() != 1) {
            *(Cor_exEnv->err_strm)
                << "Error: count_distinct(<>) aggregation: wrong arity.\n";
            return(0);
        }

        Cor_ArgList *val_listc = new Cor_ArgList(1);
        (*val_listc)[0] = val;
        Cor_Tuple *val_tuplec = new Cor_Tuple(val_listc);

        Cor_StorageRelation *drelc = old_tuple->args()[argnum].dist_rel();

        if (drelc->is_subsumed(val_tuplec)) {
           // then not a distinct element, so don't need to do anything.
           return(1);
        }

        // otherwise, need to update the grouping value, and the set

        drelc->insert_tuple(val_tuplec);

        int ival = old_tuple->args()[argnum]._agg_num->NumIval();
       
        old_tuple->args()[argnum].set_kind(AGG_RESULT_ARG);
        *(old_tuple->args()[argnum]._agg_num) = (ival+1);
        //old_tuple->args()[argnum].agg._dist_rel = drelc;

        return(1);
    }

    case SUM_DGRPING_ARG:
    case PROD_DGRPING_ARG:
    {
        if (grp_arity() != 1) {
            *(Cor_exEnv->err_strm)
                << "Error: sum_distinct(<>)  or prod_distinct(<>) "
                << "aggregation: wrong arity.\n";
            return(0);
        }

        if ( !old_tuple->args()[argnum]._agg_num->is_num()
                    || !val.is_num() ) {
            *(Cor_exEnv->err_strm)
                << "Error: sum_distinct(<>) or prod_distinct(<>) "
                << "aggregation: not a number.\n";
            return(0);
        }
                
        Cor_ArgList *val_list = new Cor_ArgList(1);
        (*val_list)[0] = val;
        Cor_Tuple *val_tuple = new Cor_Tuple(val_list);

        Cor_StorageRelation *drel = old_tuple->args()[argnum].dist_rel();
        if (drel->is_subsumed(val_tuple)) {
           // then not a distinct element, so don't need to do anything.
           return(1);
        }

        // otherwise, need to update the grouping value, and the set

        drel->insert_tuple(val_tuple);
        
        double tmp_doub;

        if (Kind() == SUM_DGRPING_ARG) {
            tmp_doub = old_tuple->args()[argnum]._agg_num->NumDval()
                    + val.NumDval();
        } else {
            tmp_doub = old_tuple->args()[argnum]._agg_num->NumDval()
                    * val.NumDval();
        }

        *(old_tuple->args()[argnum]._agg_num) = tmp_doub;
        //old_tuple->args()[argnum].agg._dist_rel = drel;

        return(1);
    }
    }
}

// *************************************************************************
size_t Cor_CArg::show_internals() const
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    size_t my_sze = sizeof(Cor_CArg);

#ifdef    DEBUG
    cerr << "sizeof(CArg): " << my_sze << endl;

    cerr << "offset of my_arg_kind: " << offsetof(Cor_CArg,my_arg_kind) <<endl;
    cerr << "sizeof(my_arg_kind): " << sizeof(CArgKind) << endl;

    cerr << "offset of _hash: " << offsetof(Cor_CArg,_hash) << endl;
    cerr << "sizeof(_hash): " << sizeof(HashVal) << endl;

    cerr << "offset of union: " << offsetof(Cor_CArg,_doub) << endl;
    cerr << "sizeof(double): " << sizeof(double) << endl;

    cerr << endl;
#endif    DEBUG
    return(my_sze);
}

// *************************************************************************
char *Cor_CArg::kind_str() const
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
#ifdef    DEBUG
    switch (my_arg_kind) {

    default:
        cerr << "Bad enum value: " << my_arg_kind << endl;
        return("default (not in switch/case!)");

    case EMPTY_ARG:
        return("EMPTY_ARG");

    case INT_ARG:
        return("INT_ARG");

    case DOUBLE_ARG:
        return("DOUBLE_ARG");

    case SYMBOL_ARG:
        return("SYMBOL_ARG");

    case TINY_SYM_ARG:
        return("TINY_SYM_ARG");

    case CHUNK_ARG:
        return("CHUNK_ARG");

    case FUNCT_ARG:
        return("FUNCT_ARG");

    case VAR_ARG:
        return("VAR_ARG");

    case REL_ARG:
        return("REL_ARG");

    case DONT_CARE_ARG:
        return("DONT_CARE_ARG");

    case PARAMETER_ARG:
        return("PARAMETER_ARG");

      // Lists
    case NIL_ARG:
        return("NIL_ARG");

    case LIST_ARG:
        return("LIST_ARG");

      // Arithmetic Expressions
    case ADD_OP_ARG:
        return("ADD_OP_ARG");

    case SUB_OP_ARG:
        return("SUB_OP_ARG");

    case MULT_OP_ARG:
        return("MULT_OP_ARG");

    case DIV_OP_ARG:
        return("DIV_OP_ARG");

    case IDIV_OP_ARG:
        return("IDIV_OP_ARG");

    case MOD_OP_ARG:
        return("MOD_OP_ARG");


    case LSHFT_OP_ARG:
        return("LSHFT_OP_ARG");

    case RSHFT_OP_ARG:
        return("RSHFT_OP_ARG");


    case BT_AND_OP_ARG:
        return("BT_AND_OP_ARG");

    case BT_OR_OP_ARG:
        return("BT_OR_OP_ARG");

    case BT_NOT_OP_ARG:
        return("BT_NOT_OP_ARG");

    case BT_XOR_OP_ARG:
        return("BT_XOR_OP_ARG");


      // The various types of Grouping Args
    case SUM_GRPING_ARG:
        return("SUM_GRPING_ARG");

    case PROD_GRPING_ARG:
        return("PROD_GRPING_ARG");

    case CNT_GRPING_ARG:
        return("CNT_GRPING_ARG");

    case AVG_GRPING_ARG:
        return("AVG_GRPING_ARG");

    case STDDEV_GRPING_ARG:
        return("STDDEV_GRPING_ARG");


    case SUM_DGRPING_ARG:
        return("SUM_DGRPING_ARG");

    case PROD_DGRPING_ARG:
        return("PROD_DGRPING_ARG");

    case CNT_DGRPING_ARG:
        return("CNT_DGRPING_ARG");

    case AVG_DGRPING_ARG:
        return("AVG_DGRPING_ARG");


    case MIN_GRPING_ARG:
        return("MIN_GRPING_ARG");

    case MAX_GRPING_ARG:
        return("MAX_GRPING_ARG");

    case SET_GRPING_ARG:
        return("SET_GRPING_ARG");

    case ANY_GRPING_ARG:
        return("ANY_GRPING_ARG");

    case NO_AGG_GRPING_ARG:
        return("NO_AGG_GRPING_ARG");


    case AGG_RESULT_ARG:
        return("AGG_RESULT_ARG");

    case CMPLX_AGG_RESULT_ARG:
        return("CMPLX_AGG_RESULT_ARG");
    }

#else
    // This is only available to assist in debugging!
    return(NULL);
#endif    // DEBUG
}

// *************************************************************************
Cor_CArg& Cor_CArg::operator= (const Cor_CArg& that)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
      // Setting something equal to itself is disallowed.
    ASSERT(this != &that);

    Cor_CArg *that2;

      // Need to increment any reference counts here
      // To handle the args pointing to the same thing
    switch (that.my_arg_kind) {

    case REL_ARG:
        if (that._rel)
            that._rel->carg_count++;
        break;

    case AGG_RESULT_ARG:
        if (that.agg._dist_rel)
            (that.agg._dist_rel->carg_count)++;
        break;

    case SYMBOL_ARG:
        *(that.sym._str_ref_cnt) += 1;
        break;

    default:
        break;
    }

    free_mem();

      // This can change for reference counted types
    my_arg_kind = that.my_arg_kind;

    switch (my_arg_kind) {

    default:
#ifdef    DEBUG
        TELL;
        cerr << "Unknown kind " << kind_str() << endl;
#endif
        CORAL_error(COR_FATAL_ERROR, "", __PRETTY_FUNCTION__,
                __FILE__, __LINE__);
        ASSERT(0);
        break;

    case EMPTY_ARG:
        init();
        break;

    case INT_ARG:
        _int = that._int;
        break;

    case DOUBLE_ARG:
        _doub = that._doub;
        break;

    case SYMBOL_ARG: {
            sym._symbol = that.sym._symbol;
            sym._str_ref_cnt = that.sym._str_ref_cnt;
            _hash = that._hash;
        }
        break;

    case TINY_SYM_ARG:
        memcpy(_symbuf, that._symbuf, sizeof(double));
        _hash = that._hash;
        break;

    case CHUNK_ARG:
        _chunk = that._chunk;
        break;

    case FUNCT_ARG:
          // Have to change to a reference, since we're
          // not sure what this item is.
        that2 = (Cor_CArg *) &that;
        func._functor = that2->func._functor->dup();
        func._args = that2->func._args->dup();
        _hash = that2->hash();
        break;

    case VAR_ARG:
        _var_num = that._var_num;
        break;

    case REL_ARG:
        _rel = that._rel;
        break;

    case DONT_CARE_ARG:
    case PARAMETER_ARG:
        break;

      // Lists
    case NIL_ARG:
        break;

    case LIST_ARG:
        that2 = (Cor_CArg *) &that;
        _hash = that2->hash();
        list._items = that.list._items->dup();
        list._num_items = that.list._num_items;
        break;

      // Arithmetic Expressions
    case ADD_OP_ARG:
    case SUB_OP_ARG:
    case MULT_OP_ARG:
    case DIV_OP_ARG:
    case IDIV_OP_ARG:
    case MOD_OP_ARG:

    case LSHFT_OP_ARG:
    case RSHFT_OP_ARG:

    case BT_AND_OP_ARG:
    case BT_OR_OP_ARG:
    case BT_NOT_OP_ARG:
    case BT_XOR_OP_ARG:
        expr._larg = that.larg()->dup();
        expr._rarg = that.rarg()->dup();
        break;

      // The various types of Grouping Args
    case SUM_GRPING_ARG:
    case PROD_GRPING_ARG:
    case CNT_GRPING_ARG:
    case AVG_GRPING_ARG:

    case SUM_DGRPING_ARG:
    case PROD_DGRPING_ARG:
    case CNT_DGRPING_ARG:
    case AVG_DGRPING_ARG:

    case MIN_GRPING_ARG:
    case MAX_GRPING_ARG:
    case SET_GRPING_ARG:
    case ANY_GRPING_ARG:
    case NO_AGG_GRPING_ARG:
        that2 = (Cor_CArg *) &that;
        grp._nested_args = that2->grp._nested_args->dup();
        grp._seq_num     = that2->grp._seq_num;
        _chk_subsum      = that2->_chk_subsum;
        break;

    case AGG_RESULT_ARG:
        that2 = (Cor_CArg *) &that;
        agg._agg_count   = that2->agg._agg_count;
        agg._dist_rel    = that2->agg._dist_rel;
        if (that2->_agg_num)
            _agg_num = that2->_agg_num->dup();
        else
            _agg_num = NULL;
        break;
    }

    return(*this);
}

// *************************************************************************
void Cor_CArg::var_list(Cor_IntStack *var_stk)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (this) {

        switch (my_arg_kind) {

        default:
#ifdef    DEBUG
            TELL;
            cerr << "Unknown kind " << kind_str() << endl;
#endif
        case EMPTY_ARG:
            break;

        case INT_ARG:
        case DOUBLE_ARG:
        case SYMBOL_ARG:
        case TINY_SYM_ARG:
        case CHUNK_ARG:
            break;

        case FUNCT_ARG:
            FOR_EACH_ARG(farg,args()) {
                farg->var_list(var_stk);
            } END_EACH_ARG;
            break;

        case VAR_ARG: 

              // Check for duplicates.
            if (! var_stk->contains(var_num()))
                var_stk->push(var_num());
            break;

        case REL_ARG:
        case DONT_CARE_ARG:
        case PARAMETER_ARG:
            break;

          // Lists
        case LIST_ARG: {
            int j;
            for (j=list._num_items-1; j >= 0; j--) {
                Cor_CArg *tmp = list._items->get_ith(j);
                tmp->var_list(var_stk);
            }
            break;
        }

        case NIL_ARG:
            break;

          // Arithmetic Expressions
        case ADD_OP_ARG:
        case SUB_OP_ARG:
        case MULT_OP_ARG:
        case DIV_OP_ARG:
        case IDIV_OP_ARG:
        case MOD_OP_ARG:

        case LSHFT_OP_ARG:
        case RSHFT_OP_ARG:

        case BT_AND_OP_ARG:
        case BT_OR_OP_ARG:
        case BT_NOT_OP_ARG:
        case BT_XOR_OP_ARG:

            expr._larg->var_list(var_stk);
            expr._rarg->var_list(var_stk);
            break;

          // The various types of Grouping Args
        case SUM_GRPING_ARG:
        case PROD_GRPING_ARG:
        case CNT_GRPING_ARG:
        case AVG_GRPING_ARG:
            break;
    
        case SUM_DGRPING_ARG:
        case PROD_DGRPING_ARG:
        case CNT_DGRPING_ARG:
        case AVG_DGRPING_ARG:
    
        case MIN_GRPING_ARG:
        case MAX_GRPING_ARG:
        case SET_GRPING_ARG:
        case ANY_GRPING_ARG:
        case NO_AGG_GRPING_ARG:
            break;
    
        case AGG_RESULT_ARG:
            break;
        }
    }
}

// *************************************************************************
Variable Cor_CArg::max_vars()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    switch (my_arg_kind) {

    default:
#ifdef    DEBUG
        TELL;
        cerr << "Unknown kind in max_vars(): " << kind_str() << endl;
#endif
    case EMPTY_ARG:
        break;

    case INT_ARG:
    case DOUBLE_ARG:
    case SYMBOL_ARG:
    case TINY_SYM_ARG:
        break;

    case CHUNK_ARG:
        break;

    case FUNCT_ARG:
        return(args().max_vars());
        break;

    case VAR_ARG:
        return(_var_num+1);
        break;

    case REL_ARG:
    case DONT_CARE_ARG:
    case PARAMETER_ARG:
        break;

      // Lists
    case LIST_ARG: {
        int j, tmp_mv, mv = 0;
        for (j=list._num_items-1; j >= 0; j--) {
            Cor_CArg *tmp = list._items->get_ith(j);
            tmp_mv = tmp->max_vars();
            if (tmp_mv > mv)
                mv = tmp_mv;
        }
        return(mv);
    }

    case NIL_ARG:
        break;

      // Arithmetic Expressions
    case ADD_OP_ARG:
    case SUB_OP_ARG:
    case MULT_OP_ARG:
    case DIV_OP_ARG:
    case IDIV_OP_ARG:
    case MOD_OP_ARG:

    case LSHFT_OP_ARG:
    case RSHFT_OP_ARG:

    case BT_AND_OP_ARG:
    case BT_OR_OP_ARG:
    case BT_NOT_OP_ARG:
    case BT_XOR_OP_ARG: {

            int rmv = 0;
            int lmv = expr._larg->max_vars();
            if (expr._rarg)
                rmv = expr._rarg->max_vars();
            return( (lmv > rmv) ? lmv : rmv );
        }
        break;

      // The various types of Grouping Args
    case SUM_GRPING_ARG:
    case PROD_GRPING_ARG:
    case CNT_GRPING_ARG:
    case AVG_GRPING_ARG:
        break;

    case SUM_DGRPING_ARG:
    case PROD_DGRPING_ARG:
    case CNT_DGRPING_ARG:
    case AVG_DGRPING_ARG:
        break;

    case MIN_GRPING_ARG:
    case MAX_GRPING_ARG:
    case SET_GRPING_ARG:
    case ANY_GRPING_ARG:
    case NO_AGG_GRPING_ARG:
        break;

    case AGG_RESULT_ARG:
        break;
    }

    return(0);
}

// *************************************************************************
int Cor_CArg::scan_vars(void *data, VarFunc *vfunc)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    switch (my_arg_kind) {

    default:
#ifdef    DEBUG
        TELL;
        cerr << "Unknown kind in scan_vars(): " << kind_str() << endl;
#endif
    case EMPTY_ARG:
        break;

    case INT_ARG:
    case DOUBLE_ARG:
    case SYMBOL_ARG:
    case TINY_SYM_ARG:
    case CHUNK_ARG:
        break;

    case FUNCT_ARG:
        FOR_EACH_ARG(farg, args()) {
            int code = farg->scan_vars(data,vfunc);
            if (code)
                return(code);
        } END_EACH_ARG;
        break;

    case VAR_ARG:
        return( (vfunc)(this,data));

    case REL_ARG:
    case DONT_CARE_ARG:
    case PARAMETER_ARG:
        break;

      // Lists
    case LIST_ARG: {
        int j, code = 0;
        for (j=list._num_items-1; j >= 0; j--) {
            Cor_CArg *tmp = list._items->get_ith(j);
            if ((code = tmp->scan_vars(data,vfunc)))
                break;
        }
        return(code);
    }

    case NIL_ARG:
        break;

      // Arithmetic Expressions
    case ADD_OP_ARG:
    case SUB_OP_ARG:
    case MULT_OP_ARG:
    case DIV_OP_ARG:
    case IDIV_OP_ARG:
    case MOD_OP_ARG:
        break;

    case LSHFT_OP_ARG:
    case RSHFT_OP_ARG:
        break;

    case BT_AND_OP_ARG:
    case BT_OR_OP_ARG:
    case BT_NOT_OP_ARG:
    case BT_XOR_OP_ARG:
        break;

      // The various types of Grouping Args
    case SUM_GRPING_ARG:
    case PROD_GRPING_ARG:
    case CNT_GRPING_ARG:
    case AVG_GRPING_ARG:
        break;

    case SUM_DGRPING_ARG:
    case PROD_DGRPING_ARG:
    case CNT_DGRPING_ARG:
    case AVG_DGRPING_ARG:
        break;

    case MIN_GRPING_ARG:
    case MAX_GRPING_ARG:
    case SET_GRPING_ARG:
    case ANY_GRPING_ARG:
    case NO_AGG_GRPING_ARG:
        break;

    case AGG_RESULT_ARG:
        break;
    }

    return(0);
}

// *************************************************************************
Cor_CArg *Cor_CArg::dup()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_CArg *ret = NULL;

    if (!this)
        return(NULL);

    switch (my_arg_kind) {

    default:
        if (this)
            ret = new Cor_CArg(*this);
        break;

    case FUNCT_ARG:

        ret = new Cor_CArg(FUNCT_ARG);
        ret->func._functor = func._functor->dup();
        ret->func._args = func._args->dup();
        ret->_hash = _hash;
        break;

    case LIST_ARG:
        // Should check if ref cnted
        ret = new Cor_CArg(*this); 
        break;
    }

    return(ret);
}

// *************************************************************************
Cor_CArg *Cor_CArg::list_dup()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    ASSERT(my_arg_kind == LIST_ARG);

    Cor_CArg *ret = new Cor_CArg( items()->dup_bottom(num_items()) ); 

    return(ret);
}

// *************************************************************************
void Cor_CArg::free_mem(void)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    switch (my_arg_kind) {

    default:
#ifdef    DEBUG
        TELL;
        cerr << "Unknown kind " << kind_str() << endl;
        return;
#endif
        break;

    case EMPTY_ARG:
        return;

    case INT_ARG:
    case DOUBLE_ARG:
        break;

    case SYMBOL_ARG:
        if ( --(*(sym._str_ref_cnt)) == 0 )
            Cor_string_ref_table->destroy(sym._symbol,&_hash);
        break;

    case TINY_SYM_ARG:
    case CHUNK_ARG:
        break;

    case FUNCT_ARG:
        delete func._functor;
        delete func._args;
        break;

    case VAR_ARG:
        break;

    case REL_ARG:
        if (_rel)
            if (--(_rel->carg_count) == 0)
                delete _rel;
        break;

    case DONT_CARE_ARG:
    case PARAMETER_ARG:
        break;

      // Lists
    case LIST_ARG:
        delete list._items;
        break;

    case NIL_ARG:
        break;


      // Arithmetic Expressions
    case ADD_OP_ARG:
    case SUB_OP_ARG:
    case MULT_OP_ARG:
    case DIV_OP_ARG:
    case IDIV_OP_ARG:
    case MOD_OP_ARG:

    case LSHFT_OP_ARG:
    case RSHFT_OP_ARG:

    case BT_AND_OP_ARG:
    case BT_OR_OP_ARG:
    case BT_NOT_OP_ARG:
    case BT_XOR_OP_ARG:
        delete expr._larg;
        delete expr._rarg;
        break;

      // The various types of Grouping Args
    case SUM_GRPING_ARG:
    case PROD_GRPING_ARG:
    case CNT_GRPING_ARG:
    case AVG_GRPING_ARG:

    case SUM_DGRPING_ARG:
    case PROD_DGRPING_ARG:
    case CNT_DGRPING_ARG:
    case AVG_DGRPING_ARG:

    case MIN_GRPING_ARG:
    case MAX_GRPING_ARG:
    case SET_GRPING_ARG:
    case ANY_GRPING_ARG:
    case NO_AGG_GRPING_ARG:
        delete grp._nested_args;
        break;

    case AGG_RESULT_ARG:
        delete _agg_num;
        if (agg._dist_rel)
            if (--(agg._dist_rel->carg_count) == 0)
                delete agg._dist_rel;
        break;
    }

#ifdef    DEBUG
    // This helps to identify problems if this memory
    // is accidentally reused.
    //init();
#endif
    my_arg_kind = EMPTY_ARG;
}

// *************************************************************************
// UnMake a list of items into a tuple.
// ('make_list()' is in the ArgList class.)
Cor_Tuple *Cor_CArg::unmake_list()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_Tuple *tuple = NULL;

    if (is_cons()) {
        int j, n = num_items();
        Cor_ArgList *args = new Cor_ArgList(n-1);

        for (j=1; j < n; j++) {
            Cor_CArg *arg = list._items->get_ith(j);
            (*args)[n-j-1] = *arg;
        }

        tuple = new Cor_Tuple(args);
    }

    return(tuple);
}

// *************************************************************************
// UnMake a list of lists
int Cor_CArg::unmake_lists(Cor_TupleStack *stk)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int gotsome = 0;

    if (is_cons()) {
        int j, n = num_items();

        for (j=1; j < n; j++) {
            Cor_CArg *arg = list._items->get_ith(j);
            Cor_Tuple *tuple = arg->unmake_list();

            if (tuple) {
                stk->push(UnknownHashValue,tuple);
                gotsome = 1;
            }
        }
    } 

    return(gotsome);
}

// *************************************************************************
// Substitute the parameters in. Return the number used.
int  Cor_CArg::substitute_params( Cor_CArgStack *params )
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif

    int cnt = 0;


    switch (my_arg_kind) {

    default:
#ifdef    DEBUG
        TELL;
        cerr << "Unknown kind " << kind_str() << endl;
#endif
    case EMPTY_ARG:
        break;

    case INT_ARG:
    case DOUBLE_ARG:
    case SYMBOL_ARG:
    case TINY_SYM_ARG:
    case CHUNK_ARG:
    case VAR_ARG:
    case DONT_CARE_ARG:
    case NIL_ARG:
        break;

    case FUNCT_ARG:
        FOR_EACH_ARG(arg, args()) {
            cnt += arg->substitute_params(params);
        } END_EACH_ARG;
        break;

    case REL_ARG:
        // Is it possible to use a parameter here?
        break;

    case PARAMETER_ARG: {
            Cor_CArg *carg = params->pop();
            *this = *carg;
            delete carg;
            cnt++; 
        }
        break;

      // Lists
    case LIST_ARG: {
            int j;
            for (j=list._num_items-1; j >= 0; j--) {

                Cor_CArg *ele = list._items->get_ith(j);
                cnt += ele->substitute_params(params); 
            }
        }
        break;

      // Arithmetic Expressions
    case ADD_OP_ARG:
    case SUB_OP_ARG:
    case MULT_OP_ARG:
    case DIV_OP_ARG:
    case IDIV_OP_ARG:
    case MOD_OP_ARG:
    case LSHFT_OP_ARG:
    case RSHFT_OP_ARG:
    case BT_AND_OP_ARG:
    case BT_OR_OP_ARG:
    case BT_XOR_OP_ARG:
    case BT_NOT_OP_ARG: {
            if (larg())
                cnt += larg()->substitute_params(params);
            if (rarg())
                cnt += rarg()->substitute_params(params);
        }
        break;


      // The various types of Grouping Args

    case SET_GRPING_ARG:
    case NO_AGG_GRPING_ARG:
    case CNT_GRPING_ARG:
    case SUM_GRPING_ARG:
    case PROD_GRPING_ARG:
    case MIN_GRPING_ARG:
    case MAX_GRPING_ARG:
    case ANY_GRPING_ARG:
    case AVG_GRPING_ARG:
    case CNT_DGRPING_ARG:
    case AVG_DGRPING_ARG:
    case PROD_DGRPING_ARG:
    case SUM_DGRPING_ARG:
    case USR_DEFD_GRPING_ARG:
        break;

    case AGG_RESULT_ARG:
    case CMPLX_AGG_RESULT_ARG:
        break;
    }

    return(cnt);
}

// *************************************************************************
