//**************************************************************************
//    ExecutionEnv.C
//**************************************************************************

#include "ExecutionEnv.h"

#include "Config.h"
#include "CoralError.h"
#include "Globals.h"
#include "BuiltinSymbols.h"
#include "Debug.h"

//**************************************************************************
Cor_ExecutionEnv::Cor_ExecutionEnv(Cor_ExecutionEnv *proto,
        CoralStream *in_file, CoralStream *out_file, CoralStream *err_file)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    explain_strm = NULL;        // non-NULL only if explain is running

    SM_initialized = proto->SM_initialized;
    C_exec_mode = proto->C_exec_mode;
    C_insert_mode_default = proto->C_insert_mode_default;
    C_quiet_mode_default = proto->C_quiet_mode_default;
    GlobalRelationOptions = proto->GlobalRelationOptions;
    dbg_unify= dbg_subsumes= dbg_indexing= dbg_aggregates= dbg_get_next= 
            proto->dbg_unify;
    dbg_sets = proto->dbg_sets;
    dbg_modules = proto->dbg_modules;
    dbg_deletes = proto->dbg_deletes;
    dbg_ordsearch = proto->dbg_ordsearch;
    dbg_pipelining = proto->dbg_pipelining;
    dbg_backtrack = proto->dbg_backtrack;
    C_hash_buckets_default = proto->C_hash_buckets_default;
    C_delta_buckets_default = proto->C_delta_buckets_default;
    C_incr_ratio_default = proto->C_incr_ratio_default;
    C_max_occupancy_default = proto->C_max_occupancy_default;
    C_max_recursion_depth_default = proto->C_max_recursion_depth_default;
    C_relation_type_default = proto->C_relation_type_default;
    C_connect_exodus_default = proto->C_connect_exodus_default;


    profile_sn = proto->profile_sn;
    profile_scc = proto->profile_scc;
    profile_module = proto->profile_module;


    if (in_file)
        in_strm = in_file;
    else
        in_strm = proto->in_strm;

    if (err_file){
        err_strm = err_file;
        profile_strm = err_file;
        trace_strm = err_file;
    } else {
        err_strm = proto->err_strm;
        profile_strm = proto->profile_strm;
        trace_strm = proto->trace_strm;
    }
    trace_filename = proto->trace_filename;
    error_filename = proto->error_filename;

    if (out_file)
        out_strm = out_file;
    else 
        out_strm = proto->out_strm;

    out_filename = proto->out_filename;
    in_filename  = proto->in_filename;
    
    C_answer_style_default = proto->C_answer_style_default;
    C_use_quotes = proto->C_use_quotes;
    C_create_magic_default = proto->C_create_magic_default;
    C_read_coralrc_default = proto->C_read_coralrc_default;
    C_verbose_answer_default = proto->C_verbose_answer_default;
    C_verbose_aggsel_default = proto->C_verbose_aggsel_default;

    C_print_prompt_default = proto->C_print_prompt_default;
    C_do_history_default = proto->C_do_history_default;

    C_preprocessing_default = proto->C_preprocessing_default;
    C_rewriting_default = proto->C_rewriting_default;
    C_use_supplementary_magic_default 
            = proto->C_use_supplementary_magic_default;
    C_use_magic_default = proto->C_use_magic_default;
    C_multisets_default = proto->C_multisets_default;
    C_check_subsum_default = proto->C_check_subsum_default;
    C_single_scc_default = proto->C_single_scc_default;
    C_single_answer_default = proto->C_single_answer_default;
    C_convert_functions_default = proto->C_convert_functions_default;
    C_index_deltas_default = proto->C_index_deltas_default;
    C_use_factor_magic_default = proto->C_use_factor_magic_default;
    C_use_pipelining_default = proto->C_use_pipelining_default;
    C_use_ordsearch_default = proto->C_use_ordsearch_default;
    C_use_lazy_eval_default = proto->C_use_lazy_eval_default;
    C_use_monotonic_default = proto->C_use_monotonic_default;
    C_interactive_mode_default = proto->C_interactive_mode_default;
    C_use_factoring_default = proto->C_use_factoring_default;
    C_diskrel_default = proto->C_diskrel_default;
    C_eager_indexing_default = proto->C_eager_indexing_default;
    C_sip_reorder_default = proto->C_sip_reorder_default;
    C_naive_bcktrking_default = proto->C_naive_bcktrking_default;
    C_nested_scc_eval_default = proto->C_nested_scc_eval_default;

    stackmark = new Cor_StackMark;

    int len = strlen(proto->coral_path);
    coral_path = new char[len+1];
    strcpy(coral_path, proto->coral_path);
    coral_path[len] = '\0';

    C_interrupt_raised = proto->C_interrupt_raised;
    quit = proto->quit;
    yield = proto->yield;

    C_at_end_module = proto->C_at_end_module;
    in_query_loop = proto->in_query_loop;
    c_module_separator = proto->c_module_separator;
}

//**************************************************************************
Cor_ExecutionEnv::Cor_ExecutionEnv()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    explain_strm = NULL;       // non-NULL only if explain is running
    init();
}

//**************************************************************************
void Cor_ExecutionEnv::init(CoralStream *in_file,
            CoralStream *out_file, CoralStream *err_file)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    SM_initialized = 0;
    C_exec_mode = COR_DOING_INTERPRET;
    C_insert_mode_default = COR_INSERT_MODE_DEFAULT;
    C_quiet_mode_default = 0;
    GlobalRelationOptions = 0;
    dbg_unify= dbg_subsumes= dbg_indexing= dbg_aggregates= dbg_get_next= 0;
    dbg_sets=0;
    dbg_modules = 0;
    dbg_deletes = 0;
    dbg_ordsearch=0;
    dbg_pipelining = 0;
    dbg_backtrack = 0;
    C_hash_buckets_default = COR_DEFAULT_INDEX_BUCKETS;
    C_delta_buckets_default = COR_DEFAULT_DELTA_BUCKETS;
    C_incr_ratio_default = COR_INDEX_INCR_RATIO;
    C_max_occupancy_default = COR_INDEX_OCCUPANCY;
    C_max_recursion_depth_default = COR_MAX_RECURSION_DEPTH;
    C_relation_type_default = COR_RELATION_TYPE_DEFAULT;
    C_connect_exodus_default = COR_CONNECT_EXODUS_DEFAULT;
    
    profile_sn = COR_PROFILE_SN_DEFAULT;
    profile_scc = COR_PROFILE_SN_DEFAULT;
    profile_module = COR_PROFILE_MODULE_DEFAULT;

    if (in_file)
        in_strm = in_file;
    else
        in_strm = cin_strm;

    if (err_file){
        err_strm = err_file;
    } else {
        err_strm = cerr_strm;
    }

    profile_strm = err_strm;
    trace_strm = err_strm;

    error_filename = "stderr";
    profile_filename = "stderr";
    trace_filename = "stderr";

    if (out_file)
        out_strm = out_file;
    else
        out_strm = cout_strm;

    in_filename  = "stdin";
    out_filename = "stdout";
    
    C_answer_style_default = COR_ANSWER_STYLE_DEFAULT;
    C_use_quotes = COR_USE_QUOTES_DEFAULT;
    C_create_magic_default = COR_CREATE_MAGIC_DEFAULT;
    C_read_coralrc_default = COR_READ_CORALRC_DEFAULT;
    C_verbose_answer_default = COR_VERBOSE_ANSWER_DEFAULT;
    C_verbose_aggsel_default = COR_VERBOSE_AGGSEL_DEFAULT;
    C_print_prompt_default = COR_PRINT_PROMPT_DEFAULT;
    C_do_history_default = COR_DO_HISTORY_DEFAULT;

    C_preprocessing_default = COR_PREPROCESSING_DEFAULT;
    C_rewriting_default = COR_REWRITING_DEFAULT;
    C_use_supplementary_magic_default = COR_USE_SUPPLEMENTARY_MAGIC_DEFAULT ;
    C_use_magic_default = COR_USE_MAGIC_DEFAULT;
    C_multisets_default = COR_MULTISETS_DEFAULT;
    C_check_subsum_default = COR_CHECK_SUBSUM_DEFAULT;
    C_single_scc_default = COR_SINGLE_SCC_DEFAULT;
    C_single_answer_default = COR_SINGLE_ANSWER_DEFAULT;
    C_convert_functions_default = COR_CONVERT_FUNCTIONS_DEFAULT;
    C_index_deltas_default = COR_INDEX_DELTA_DEFAULT;
    C_use_factor_magic_default = COR_USE_FACTOR_MAGIC_DEFAULT;
    C_use_pipelining_default = COR_USE_PIPELINING_DEFAULT;
    C_use_ordsearch_default = COR_USE_ORDSEARCH_DEFAULT;
    C_use_lazy_eval_default = COR_USE_LAZY_EVAL_DEFAULT;
    C_use_monotonic_default = COR_USE_MONOTONIC_DEFAULT;
    C_interactive_mode_default = COR_INTERACTIVE_MODE_DEFAULT;
    C_use_factoring_default = COR_USE_FACTORING_DEFAULT;
    C_diskrel_default = COR_DISKREL_DEFAULT;
    C_eager_indexing_default = COR_EAGER_INDEXING_DEFAULT;
    C_sip_reorder_default = COR_SIP_REORDER_DEFAULT;
    C_naive_bcktrking_default = COR_NAIVE_BACKTRACKING_DEFAULT;
    C_nested_scc_eval_default = COR_NESTED_SCC_EVAL_DEFAULT;
    
    stackmark = new Cor_StackMark;

    char *cr = getenv("CORALROOT");

    if (!cr) {
#        ifdef CORALROOT
            cr = CORALROOT;
#        else
            *err_strm
                << "Warning! CORALROOT not set in environment!\n";
            cr = DEFAULT_CORAL_PATH;
            *err_strm << " Using " << DEFAULT_CORAL_PATH << endl;
#        endif
    }

    int len = strlen(cr);
    coral_path = new char[len+1];
    strcpy(coral_path,cr);
    coral_path[len] = '\0';

    C_interrupt_raised = 0;
    quit = 0;
    yield = 0;

    C_at_end_module = 1;
    in_query_loop = 1;
    c_module_separator = 0;
}

//**************************************************************************
Cor_ExecutionEnv::~Cor_ExecutionEnv()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    delete stackmark;
    delete [] coral_path;
}

//**************************************************************************
void Cor_ExecutionEnv::flush_all_outputs()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (out_strm)
        *out_strm << flush;
    if (err_strm)
        *err_strm << flush;
    if (trace_strm)     
        *trace_strm << flush;
    if (profile_strm)
        *profile_strm << flush;

#ifdef    DO_EXPLAIN
    if (explain_strm)
        *explain_strm << flush;
#endif    

}

//**************************************************************************
void Cor_ExecutionEnv::display(CoralStream *outf)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    *outf << "\tmode :\t\t\t";

    switch (C_exec_mode) {
    case COR_DOING_INTERPRET:
        *outf << "INTERPRETER\n";
        break;
    case COR_DOING_COMPILE:
        *outf << "COMPILER\n";
        break;
    case COR_DOING_MAGIC:
        *outf << "MAGIC Rewriting\n";
        break;
    case COR_DOING_FACTOR:
        *outf << "FACTOR Rewriting\n";
        break;
    case COR_DOING_IMPERATIVE:
        *outf << "Imperative module\n";
        break;
    case COR_DOING_FACTOR_MAGIC:
    default:
        *cerr_strm << "This can't happen\n";
        break;
    }

    *outf << "\tinsert_mode: \t\t"
          << on_off_str(C_insert_mode_default) << endl;
    *outf << "\tquiet_mode : \t\t"
          << on_off_str(C_quiet_mode_default) << endl;

    *outf << "--- Defaults for Rewriting ---\n\n";

    *outf << "\trewriting : \t\t"
          << on_off_str(C_rewriting_default) << endl;
    *outf << "\tsup_magic :\t\t"
          << on_off_str(C_use_supplementary_magic_default) << endl;
    *outf << "\tmagic :\t\t\t"
          << on_off_str(C_use_magic_default) << endl;
    *outf << "\tfactoring : \t\t"
          << on_off_str(C_use_factoring_default) << endl << endl;

    *outf << "--- Defaults for Execution ---\n\n";

    *outf << "\tinteractive_mode :\t"
          << on_off_str(C_interactive_mode_default) << endl;
    *outf << "\t" << Cor_UsePipeliningSymbol->symbol() << " :\t\t"
          << on_off_str(C_use_pipelining_default) << endl;
    *outf << "\t" << Cor_UseOrdSearchSymbol->symbol() << " :\t"
          << on_off_str(C_use_ordsearch_default) << endl;
    *outf << "\t" << Cor_LazyEvalSymbol->symbol() << " :\t\t"
          << on_off_str(C_use_lazy_eval_default) << endl;
    *outf << "\t" << Cor_MonotonicSymbol->symbol() << " :\t\t"
          << on_off_str(C_use_monotonic_default) << endl;

    *outf << "\t" << Cor_MultisetSymbol->symbol() << " :\t\t"
          << on_off_str(C_multisets_default) << endl;
    *outf << "\t" << Cor_CheckSubsumSymbol->symbol() << " :\t"
          << on_off_str(C_check_subsum_default) << endl;
    *outf << "\t" << Cor_IndexDeltaSymbol->symbol() << " :\t\t"
          << on_off_str(C_index_deltas_default) << endl;
    *outf << "\t" << Cor_DiskResidentSymbol->symbol() << " :\t\t"
          << on_off_str(C_diskrel_default) << endl;
    *outf << "\t" << Cor_EagerIndexingSymbol->symbol() << " :\t"
          << on_off_str(C_eager_indexing_default) << endl;
    *outf << "\t" << Cor_SipReorderSymbol->symbol() << " :\t\t"
          << on_off_str(C_sip_reorder_default) << endl;
    *outf << "\t" << Cor_NaiveBacktrackSymbol->symbol() << " :\t"
          << on_off_str(C_naive_bcktrking_default) << endl;
    *outf << "\t" << Cor_NestedSccEvalSymbol->symbol() << " :\t"
          << on_off_str(C_nested_scc_eval_default) << endl;

    *outf << "\n--- Defaults for Output ---\n";

    *outf << "\tout_strm :\t\t" << out_filename << endl;
    *outf << "\ttrace_strm :\t\t" << trace_filename << endl;
    *outf << "\ttracing :\t\t"
          << on_off_str(
                   Cor_exEnv->GlobalRelationOptions & REL_DISPLAY_INSERTIONS)
                << endl;
    *outf << "\tprofiling :\t\t"
          << on_off_str(Cor_exEnv->GlobalRelationOptions & REL_PROFILE_STATS)
          << endl;
    *outf << "\texplain :\t\t"
          << on_off_str(Cor_exEnv->GlobalRelationOptions & REL_EXPLAIN) << endl;


    *outf << "\tanswer_style :\t";
    switch (C_answer_style_default) {

    case COR_TUPLE_ANSWER_PRINT:
        *outf << "\ttuple\n";
        break;

    case COR_BINDING_ANSWER_PRINT:
        *outf << "\tbinding\n";
        break;

    case COR_EMPTY_ANSWER_PRINT:
        *outf << "\tempty\n";
        break;

    case COR_LITERAL_ANSWER_PRINT:
        *outf << "\tliteral\n";
        break;

    default:
        *outf << "\t-unknown-\n";
        break;
    }

    *outf << "\tverbose_answer :\t"
          << on_off_str(C_verbose_answer_default) << endl;
    *outf << "\tprint_prompt :\t\t"
          << on_off_str(C_print_prompt_default) << endl;
    *outf << "\tcreate_magic :\t\t"
          << on_off_str(C_create_magic_default) << endl;
    *outf << "\thistory :\t\t"
          << on_off_str(C_do_history_default) << endl;

    *outf << "\n--- Defaults for Profiling ---\n";

    *outf << "\tprofile_module :\t"
          << on_off_str(profile_module) << endl;
    *outf << "\tprofile_scc :\t\t"
          << on_off_str(profile_scc) << endl;
    *outf << "\tprofile_sn :\t\t"
          << on_off_str(profile_sn) << endl;

#    ifdef DEBUG
        *outf << "\n--- Defaults at System Level ---\n";

        *outf << "\tconnect_exodus: \t"
              << on_off_str(C_connect_exodus_default) << endl;
        *outf << "\tpreprocessing : \t"
              << on_off_str(C_preprocessing_default) << endl;
        *outf << "\t" << Cor_SingleSccSymbol->symbol()
              << " :\t\t" << on_off_str(C_single_scc_default) << endl;
        *outf << "\t" << Cor_ConvertFunctionsSymbol->symbol()
              << " :\t" << on_off_str(C_convert_functions_default) << endl;

        *outf << "\n--- Defaults for System Debugging ---\n";

        *outf << "\tdebug_unify :\t\t"
              << on_off_str(dbg_unify) << endl;
        *outf << "\tdebug_subsumes :\t"
              << on_off_str(dbg_subsumes) << endl;
        *outf << "\tdebug_indexing :\t"
              << on_off_str(dbg_indexing) << endl;
        *outf << "\tdebug_aggregates :\t"
              << on_off_str(dbg_aggregates) << endl;
        *outf << "\tdebug_sets :\t\t"
              << on_off_str(dbg_sets) << endl;
        *outf << "\tdebug_get_next :\t"
              << on_off_str(dbg_get_next) << endl;
        *outf << "\tdebug_ordered_search :\t"
              << on_off_str(dbg_ordsearch) << endl;
        *outf << "\tdebug_pipelining :\t"
              << on_off_str(dbg_pipelining) << endl;
        *outf << "\tdebug_backtrack :\t"
              << on_off_str(dbg_backtrack) << endl;


        *outf << "\n--- Defaults for System Tuning ---\n";

        *outf << "\thash_buckets :\t\t" << C_hash_buckets_default << endl;
        *outf << "\tdelta_buckets :\t\t" << C_delta_buckets_default << endl;
        *outf << "\tincr_ratio :\t\t" << C_incr_ratio_default << endl;
        *outf << "\tmax_occupancy :\t\t" << C_max_occupancy_default << endl;
        *outf << "\trecursion_depth :\t"
              << C_max_recursion_depth_default << endl;
        *outf << "\trelation_type :\t\t" << C_relation_type_default << endl;
#    endif 
}

//**************************************************************************
