// *************************************************************************
//    Pipelined.C
// *************************************************************************

#include "Pipelined.h"

#include "QFModuleData.h"
#include "QFPipelinedModuleData.h"
#include "RuleInfo.h"
#include "CompileRule.h"
#include "ExportInfo.h"
#include "Literal.h"
#include "Compile.h"
#include "ParserStruct.h"
#include "DerivedMethod.h"
#include "Globals.h"
#include "Debug.h"

// *************************************************************************

static void init_module_preds(Cor_QFModuleData *md);

// *************************************************************************
static void init_module_preds(Cor_QFModuleData *md)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_ModulePred *predicates = md->predicates->next;

    for (; predicates; predicates = predicates->next) { 

        Cor_CArg name = predicates->pred->name;
        int arity = predicates->pred->arity;
        predicates->offset = Cor_get_offset(md->F, name, arity);

        switch (predicates->kind) {
        case QueryPredKind:
            predicates->check_subsum = Cor_determine_subsum_status(md, name,
                md->module_info->CheckSubsum);
            break;
        case LocalPredKind:
            predicates->check_subsum = Cor_determine_subsum_status(md, name,
                md->module_info->CheckSubsum);
            break;
        case MagicQueryPredKind:
            predicates->check_subsum = Cor_determine_subsum_status(md, name,
                md->module_info->CheckSubsum);
            break;
        case ExternalPredKind:
            break;
        }
        Cor_get_annos(md, predicates);
    }
}

// *************************************************************************
//  This function is passed a pointer to the QFPipelinedModuleData 
//  the information on the exported predicate.  This function basically
//  does for a Pipelined QFModuleData what Scc analysis does for a
//  Materialized one.
// *************************************************************************
void Cor_interpretPipelined(Cor_QFPipelinedModuleData *md,
        Cor_ExportInfo *export_info, Cor_ParserStruct *Cor_parserStruct)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int i;
    Cor_Clause *tmp;
    int tvars,tliterals; // temporary vars used for calculating maximum vals

    Cor_ModulePred *predicates = md->predicates;

    Cor_CArg exported_name = export_info->pred_name;
    Cor_Predicate *pred = export_info->predicate; /* ptr to exported pred */
    
    md->numPreds = 0;
    Cor_Predicate *p;
    for(p = Cor_parserStruct->AllPredicates;
                p!= NULL; p = p->allprednext)
    {
        Cor_Clause *c;
        for (c = p->clauses;c != *p->clause_tail; c = c->pnext)
        {
            for (int ipred = 0; ipred < c->num_preds; ipred++) 
            {
                // not using Sccs for pipelining anyway
                c->preds[ipred]->predicate->work = NULL;
                c->preds[ipred]->predicate->scc = NULL;
        
                // know whether already added using add_pred_extern() or not
                c->preds[ipred]->predicate->mark = 0;
            }
        }
        md->numPreds++;
    }
    
    // Make sure that all Relations used by the query are initialized
    // by initRulePointers. Use add_pred_extern to enter them into
    // the predicates collection to keep track of them.
    // First enter the the query predicate
    
    Cor_add_pred_extern(pred, QueryPredKind, predicates);
    // Have seen the exported predicate - all rules for it have been linked in
    pred->mark = 1;
    
    md->pred_ptrs      = new Cor_Predicate *[md->numPreds];
    md->indices        = new int*[md->ruleArray->num_rules()];
    md->maxNumVars     = new int[md->numPreds];
    md->maxNumLiterals = new int[md->numPreds];
    
    for (i = 0,p = Cor_parserStruct->AllPredicates; p != NULL;
            p= p->allprednext,i++)
    {
        md->maxNumVars[i] = md->maxNumLiterals[i] = 0; 
        md->pred_ptrs[i] = p;
        for (Cor_Clause *c = p->clauses;c != NULL ; c = c->pnext) 
        {
            tliterals=((md->ruleArray)->rInfoPtrs[c->rule_number])->
                num_literals;
            tvars =((md->ruleArray)->rInfoPtrs[c->rule_number])->num_vars;
            // create indices
            md->indices[c->rule_number] = new int[((md->ruleArray)->
                rInfoPtrs[c->rule_number])->num_literals];
            if (tliterals > md->maxNumLiterals[i])
                md->maxNumLiterals[i] = tliterals;
            if (tvars > md->maxNumVars[i])
                md->maxNumVars[i] = tvars;

            // Iterate over all the Literals that are invoked in c
            for (int ipred = 0; ipred < c->num_preds; ipred++) 
            {
                Cor_Literal *child_literal = c->preds[ipred];
                Cor_Predicate *child = child_literal->predicate;
                // If we have already seen this predicate then don't insert it 
                if (child->mark == 0)
                {
                    child->mark = 1;
                    // add this predicate too
                    tmp = child->clauses;
                    // If child->clauses == *child->clause_tail, then
                    // the child has no clauses, and is an ExternalPredKind.
                    Cor_add_pred_extern(child, tmp != *child->clause_tail ?
                        LocalPredKind : ExternalPredKind, predicates);
                }
            }
        }
    }
    
    md->clean_cache();
    md->export_pred = export_info->predicate;
    md->module_info = Cor_parserStruct->CurModule;
    Cor_parserStruct->CurModule = new Cor_ModuleInfo();

    // move some of the run time work from initRulePointers to
    // compile time here. Determine what flags need to be set, etc.
    init_module_preds(md);
    
    // Insert the module structure along with its associated exported
    // query form into the DerivedMethod table.
    if (! exported_name.is_empty()) {

        // The method really does not matter - not used anyway
        Cor_DerivedMethod *query_method = new Cor_DerivedMethod(exported_name,
            *(export_info->adornment), NULL, (Cor_QFModuleData *)md);

        if (query_method)    // Eliminate "unused" warning
            ;    // Nothing
    }
}

// *************************************************************************

