// *************************************************************************
//    Collector.C
// *************************************************************************

#include "Collector.h"
#include "Util.h"
#include "Interface.h"
#include "Globals.h"
#include "BuiltinSymbols.h"
#include "TupleIterator.h"
#include "ClientProtocol.h"

// *************************************************************************
//
// A Collector keeps track of the server-side of a single relation that
// is being solved by distributed evaluation with multiple Corals. (a reef)
//
// *************************************************************************

#define   DEFAULT_SZE        8

// *************************************************************************
Cor_CollectUser::Cor_CollectUser(Cor_User *_user, Cor_CU_Type cutype)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    user = _user;
    type = cutype;
}

// *************************************************************************
Cor_Collector::Cor_Collector(Cor_Relation *rel)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    datarel = rel;

    length = 0;
    num_users = 0;
    num_producers = 0;
    num_consumers = 0;

    if (rel)
        qenv = new Cor_BindEnv(rel->arity());
    else
        qenv = NULL;

    userArray = NULL;
}

// *************************************************************************
Cor_Collector::~Cor_Collector()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    delete qenv;
    delete [] userArray;
}

// *************************************************************************
int Cor_Collector::Find_User(Cor_User *user)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int j;

    for (j=0; j < num_users; j++)
        if (userArray[j]->user == user)
            return(j);

    return(-1);    // Not found
}

// *************************************************************************
void Cor_Collector::AddQuery(Cor_User *user, Cor_Tuple *query)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int uid;

    if ((uid = Find_User(user)) >= 0) {

        // Give them all data that we currently have
        Cor_Tuple *data;
        Cor_TupleIterator *it
            = new Cor_TupleIterator(datarel, query->args(), qenv);

        if (user->qid)
            *(Cor_exEnv->out_strm) << CP_QID << user->qid << endl;

        for (data = it->get_next_tuple(); !it->no_match(); 
                                    data = it->get_next_tuple()) {
            if (data)
                *(Cor_exEnv->out_strm) << datarel->name
                            << "(" << *data << ").\n";
        }

        *(Cor_exEnv->out_strm) << flush;
        delete it;

        // Save the query (if it's not done yet)
        Cor_QueryInfo *qi = new Cor_QueryInfo();
        qi->qid = user->qid;
        qi->query = query->_args;
        query->set_no_args();    // So the _args are not deleted
        userArray[uid]->queries.push(qi);
    }
}

// *************************************************************************
void Cor_Collector::AddedFact(Cor_User *user, Cor_Tuple *fact)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (num_users > 1) {
        for (int j=0; j < num_users; j++)
            if ((userArray[j]->user != user)
                        && !userArray[j]->queries.is_empty()) {

                Cor_QueryStack *queries = &(userArray[j]->queries);
                ostream *out = userArray[j]->user->Cor_exEnv.out_strm;

                for (int k=0; k < queries->count(); k++) {

                    Cor_QueryInfo *qi = queries->get_ith(k);
                    qenv->clear_all();

                    // Not sure about this - SPF
                    if (Cor_subsumes_arg_lists(*(qi->query),
                                    fact->args(), qenv, NULL)) {
                        if (qi->qid)
                           *out << CP_QID << qi->qid << endl; 

                        *out << datarel->name
                             << "(" << *fact << ").\n" << flush;
                    }
                }
            }
    }
}

// *************************************************************************
int  Cor_Collector::AddUser(Cor_User *user, Cor_CU_Type cutype)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    user->collector = this;

    if (num_users >= length) {

        Cor_CollectUser **tempArray = userArray;

        if (length == 0)
            length = DEFAULT_SZE;
        else
            length += length;

        userArray = new Cor_CollectUser* [length];

        for (int j=0; j < num_users; j++)
            userArray[j] = tempArray[j];

        delete [] tempArray; 
    }

    userArray[num_users++] = new Cor_CollectUser(user,cutype);

    switch (cutype) {

    default:
    case Cor_Producer:
        num_producers++;
        break;

    case Cor_Consumer:
        num_consumers++;
        break;
    }

    return(num_users-1);
}

// *************************************************************************
void Cor_Collector::DeleteUser(Cor_User *user)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int index = Find_User(user);

    *(Cor_exEnv->err_strm)
        << "Leaving collection of " << datarel->name << ".\n" << flush;

    if (index >= 0) {

        switch (userArray[index]->type) {
    
        default:
        case Cor_Producer:
            num_producers--;
            break;
    
        case Cor_Consumer:
            num_consumers--;
            break;
        }

        delete userArray[index];

          // Eliminate the hole
        for (int j=index; j < num_users; j++)
            userArray[j] = userArray[j+1];

        userArray[--num_users] = NULL;

        user->collector = NULL;
    }
}

// *************************************************************************
int  Cor_Collector::ProcessUser(Cor_User *user)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    extern Cor_CArg *Cor_retCArg;
    extern Cor_CArg *Cor_RelName;
    extern int       Cor_isquery;

    int til_yield = INSERTS_BEFORE_YIELD;

    // Get input from this user, and process it.

    if (Cor_retCArg) {
        Cor_retCArg->destroy();
        Cor_retCArg = NULL;
    }

    if (Cor_RelName) {
        Cor_RelName->destroy();
        Cor_RelName = NULL;
    }

    Cor_Tuple *tuple = NULL;

    Cor_CArg relnme;
    int rel_arity = 0;
    int done = 0;

    tuple = Cor_parseFact(NULL, relnme, &rel_arity);

    while (!done) {

        if (relnme.equals(Cor_QuitSymbol)) {
            DeleteUser(user);
            done = 1;

        } else if (relnme.equals(Cor_YieldSymbol)) {
            done = 1;

        } else if (relnme.equals(Cor_SetQidSymbol)) {
            user->qid = make_int(tuple->args()[0]);

        } else if (Cor_isquery) {
            if (relnme.equals(rel()->name)) {
                // Query the collected relation
                AddQuery(user,tuple);
            } else {
                *(Cor_exEnv->err_strm)
                    << "Error: Attempted query of a non-collected relation.\n";
                *(Cor_exEnv->err_strm)
                    << "Query was ?" << relnme << "(" << *tuple << ").\n";
            }

        } else if (tuple) {
            if (relnme.equals(rel()->name)) {
                // Insert fact into collected relation
                if (! rel()->insert_tuple(tuple))
                    delete tuple;
                else
                    AddedFact(user,tuple);

                tuple = NULL;
            } else {
                *(Cor_exEnv->err_strm)
                  << "Error: Attempted insert into a non-collected relation.\n";
                *(Cor_exEnv->err_strm)
                    << "Insert was " << relnme << "(" << *tuple << ").\n";
            }

        } else {
            // EOF or error
            DeleteUser(user);
            done = 1;
        }

        delete tuple;

          // Give someone else a chance now and then
        if (til_yield-- <= 0)
            done = 1;

    /*
        // If its a new relation find or create it
        if ((rel_arity != prev_arity) || (!relnme.equals(prev_relnme))) {

            rel = Cor_find_local_or_external_relation(relnme, rel_arity);
            prev_relnme = relnme;
            prev_arity = rel_arity;
        }

        // Check for completely ground (ie: valid) at the top-level
        for (j=0; j < rel_arity; j++) {
            if ((*(tuple->_args))[j].is_var()) {
                *(Cor_exEnv->err_strm) << "ReadTable: invalid fact: "
                                       << *tuple << " ignored.\n";
                delete tuple;
                break;
            }
        }

        if (j == rel_arity) {

        // Insert fact into Relation rel
        if (! rel->insert_tuple(tuple))
            delete tuple;
        }

        // Can't delete tuple, as the relation stores it
        // (and not just the args)

        if ((++cnt % NOT_DEAD) == 0) {
            *(Cor_exEnv->out_strm) << cnt << endl;
        }
*/
        if (!done)
            tuple = Cor_parseFact(NULL, relnme, &rel_arity);
    }
    return(0);
}

#undef  DEFAULT_SZE
// *************************************************************************
