// *************************************************************************
//    Relation.h
// *************************************************************************

#ifndef _Relation_H_
#define _Relation_H_

#include "RelationMisc.h"

#include "AggSelInfo.h"
#include "PredAnnotations.h"
#include "RMark.h"
#include "Tuple.h"
#include "Debug.h"

#ifdef ASYN_SOCK_REL
#include "Table.h"
#include "Predicate.h"
#include "SockStream.h"
#endif

// *************************************************************************

class Cor_StorageRelation;
class Cor_PrioritizeInfo;
class Cor_TupleIterator;
class Cor_GenericIndexSet;
class Cor_Collector;

// *************************************************************************

class Cor_Relation {

  protected:
    int _arity;

  public:

    Cor_CArg name;
    RelationOptions local_options; 

      // If set, check for subsumption when inserting tuples. 
    char check_subsum;
    
    char delta_indexed;
    char single_success;
    char has_release;    // specifies if the relation has a release() method

    int count;                          // number of facts in relation
    unsigned int ref_count;             // how many databases it's open in
    unsigned int scan_count;            // how many scans are open on it
    unsigned int carg_count;            // how many CArg's point to it
    int          pl_local;              // Is this a "local" pipelined rel

    HashVal _hash;                      // sufficient for inequality testing
    HashVal _eq_id;                     // sufficient for equality testing
    Cor_AggSelInfo *agg_sel_info;       // Info on aggregte selections on reln

    RelationKind r_kind;

    char         _borrow;         // Indicates the tuples belong to another
                                  // relation (we don't delete them, just
                                  // our references to them.)

    Cor_Collector  *collector;    // Our collector if we have one.

#ifdef ASYN_SOCK_REL
    int         _delta_done;      // Set if we are not expecting any
                                  // more facts from outside this program.

    Cor_RMark  *magic_mark[2];    // RMarks for the magic relation.

    int         first_send_delta; // Is the send_delta called the first time?

    CoralStream *cs;              // The socket connection for asyn_sock_rel.

    int         open_queries;     // The num of queries waiting to finish
                                  // for this relation. (needing an EOQ_:)
#endif
    
      // Ordered Search Fields
    short ordsearch;

      // used for subsumption checking in ordered search
    Cor_StorageRelation *for_ordsearch;

      // non-NULL only if (r_kind == COR_R_MAGIC) && (ordsearch)
    Cor_StorageRelation *mp_done_reln;

    void incr_ref() {}
    virtual void destroy() {}

      // To set and test the "borrow" flag
    int borrow(int bflg = 1) { return(_borrow = bflg); }
    int is_borrow()          { return(_borrow); }

      // To set and test the local pipelined flag
    int set_localpiped(int plflg = 1) { return(pl_local = plflg); }
    int is_localpiped()          { return(pl_local); }

    // Aggregate selection related methods
    virtual void add_agg_sel_anno(Cor_PredAnnotations *) {}
    virtual void add_agg_sel_info(Cor_AggSelInfo *asi) { delete asi; }

    virtual Cor_GenericIndexSet *get_index_set() {
        return(NULL);
    }

    // abstract relation methods
    int arity() const { return _arity; }
    virtual void print(Cor_BindEnv*, ostream *file) const;
    virtual void print(Cor_BindEnv*, ostream *file, char *) const;
    virtual void sprint(char *str, int *pos, Cor_BindEnv *context = NULL) const;

    virtual void print_name(ostream *file);
    virtual int isConstant() { return(1); }
    virtual int tuple_count() { return(count); }
    void printon(ostream *file) const;
    void printon(ostream *file, char *) const;
    virtual void print_facts(ostream *file,
        Cor_ArgList *arglist = NULL, int as_table=0) const;

    // get_next_tuple(): returns tuple if there is one (Builtin rels etc
    // return NULL).  Returns status via tupleit->bindenv->no_match()
    virtual Cor_Tuple *get_next_tuple(Cor_TupleIterator&) = 0;
    virtual Cor_BindEnv *get_next(Cor_TupleIterator&) = 0;

    virtual int insert_tuple(Cor_Tuple *tuple) = 0;
    virtual int insert_new(Cor_ArgList& args, Cor_BindEnv *env)=0;

    inline int insert_new(Cor_Tuple *tuple)
    {
        return insert_new(tuple->args(), NULL);
    }

      // Empty tuples out of the relation.  Delete the tuples
      // only if the deleteTuples arg is not 0
    virtual void empty_relation(int deleteTuples = 0) {
        deleteTuples = 0; /*dummy*/
    }

    virtual int tuple_delete(Cor_Tuple *tuple) {
        delete_tuple(this, tuple); 
        return(1);
    }

    virtual int tuple_insert(Cor_Tuple *tuple) {
        return this->insert_new(tuple->args(), NULL);
    }

    int tuple_update(Cor_Tuple *tuple_old, Cor_Tuple *tuple_new) {
        update_tuple(this, tuple_old, tuple_new); 
        return(1);
    }

      // Release state associated with tuple iterator
    virtual void release(Cor_TupleIterator&) { }

    virtual int add_index(Cor_ArgList *)                     { return(0); }
    virtual int add_index(Cor_ArgList *, int , Cor_ArgList *){ return(0); }
    virtual Cor_RMark *getMark()                             { return(NULL); }
    virtual void freeMark(Cor_RMark *)                       { return; }

    // THESE SHOULD BE IN STORAGERELATION !!!
    virtual HashVal hash(Cor_BindEnv *env = NULL);
    // compute a hash value for a nested set
    virtual int equals(Cor_Relation *that);
    virtual int raw_equals(Cor_Relation *rel);
    virtual int contained_in(Cor_Relation *rel);
    virtual unsigned int cardinality(Cor_Tuple *tuple);

#ifdef ASYN_SOCK_REL
    void send_delta(Cor_QFMaterializedModuleData *md, Cor_Table* table, int scc_id, int num_pred, Cor_Predicate *pred);
#endif
    
    Cor_Relation( int arity, int delta_indexed = -1 );
    Cor_Relation();

    virtual ~Cor_Relation();

    virtual void show_internals(ostream&) { }
};

// *************************************************************************

#endif _Relation_H_
