// *************************************************************************
//    StorageManager.C
// *************************************************************************

#ifdef WITH_PERSISTENCE


#include "Globals.h"
#include "CoralError.h"
#include "SMUtil.h"
#include "sm_client.h"
#include "Config.h"
#include "Schema.h"
#include "IndexStruct.h"
#include "BPlusTreeIndex.h"
#include "FileDescStruct.h"
#include "BufferGroup.h"
#include "Transaction.h"
#include "Catalog.h"
#include "StorageManager.h"
#include "BuiltinRelation.h"
#include "Debug.h"


// *************************************************************************

// Defaults for Exodus volumes 
#define DATA_VOL 8000
#define TEMP_VOL 7000

// *************************************************************************

Cor_StorageManager SM;
int verboseError;

int SM_initialized = 0;

extern void propagateChangesToDisk();

// *************************************************************************
void Cor_StorageManager::init()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int e;

    SM_initialized = 0;
    verboseError = !Cor_exEnv->C_quiet_mode_default;

    if ((Cor_exEnv->C_exec_mode == COR_DOING_FACTOR_MAGIC) ||
        (Cor_exEnv->C_exec_mode == COR_DOING_MAGIC) ||
        (Cor_exEnv->C_exec_mode == COR_DOING_FACTOR))
            return;

    // Initialize the storage manager

    char *error_msg;
    int bufsize = -1;

    e = sm_ReadConfigFile(NULL, NULL, &error_msg);
    if (e >= 0) {
        e = sm_GetClientOption("bufpages", (void *)(&bufsize));
        if (ErrorCheck(e, "sm_GetClientOption"))
            return;
    }

    if (bufsize < 0 ) {
        // Use default config values
        bufsize = COR_CLIENT_BUF_SIZE;
        e = sm_SetClientOption("bufpages", (void *)(&bufsize), SM_int);
        if (ErrorCheck(e, "sm_SetClientOption"))
            return;
    }

    e = sm_Initialize();

    if (ErrorCheck(e, "sm_Initialize"))
        return;

    //    Get the disk name
    char *volidName = getenv("EVOLID");
    // Default disk name is 'coraldisk'
    if (volidName)
        _volid = (VOLID)(atoi(volidName));
    else
        _volid = DATA_VOL;

    char *tempvolName = getenv("ETEMPVOLID");
    if (tempvolName)
        _temp_volid = (VOLID)(atoi(tempvolName));
    else
        _temp_volid = TEMP_VOL;

    //    Mount the volumes.
    e = sm_MountVolume(_volid);
    if (ErrorCheck(e, "sm_MountVolume")) {
        *(Cor_exEnv->err_strm) << "Volid of volume : " <<  _volid << "\n";
        return;
    }
    e = sm_MountVolume(_temp_volid);
    if (ErrorCheck(e, "sm_MountVolume")) {
        *(Cor_exEnv->err_strm) << "Volid of volume : " <<  _temp_volid << "\n";
        return;
    }

    //    Open a buffer group.
    if (buf.open(COR_SM_BUF_SIZE, BF_MRU))
        return;

    SM_initialized = 1;
    
    startTrans();

    if (catalog.init("CORAL_CATALOG_FILE", "CORAL_CATALOG_INDEX")) {
        SM_initialized = 0;
        return;
    }

    commitTrans();
}

// *************************************************************************
void Cor_StorageManager::shutdown()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (!Cor_exEnv->C_connect_exodus_default
        || !SM_initialized)
        return;

        // This gives inaccurate counts for adds done in aborted transactions
        // but it is at least a reasonable estimate - SPF
    startTrans();
    propagateChangesToDisk();
    commitTrans();

    if (SM_initialized) {
        sm_AbortTransaction(trans.tid);
        sm_CloseBufferGroup(buf.bufGroup);
        sm_DismountVolume(_volid);
        sm_ShutDown();
        SM_initialized = 0;
    }
}

// *************************************************************************
/*
    Retrieves an object from the storage manager.
    The var parameter is a pointer which, when the function returns,
    points to the retrieved object in the buffer pool. 

    The retrieved object is pinned in the buffer pool
    and will have to be explicitly unpinned using an sm_release() call.
    The object should be accessed only INDIRECTLY via the 'basePtr'
    field of the USERDESC structure, because EXODUS may relocate the
    contents of the buffer pool. Also note that the object is now
    PINNED in the buffer pool, and has to be unpinned using the function
    releaseObject().
*/

int Cor_StorageManager::retrieveObject(OID *oid, USERDESC ** u_desc)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (!SM_initialized) return 0;

    int e = sm_ReadObject(bufGroup(), oid, 0, READ_ALL, u_desc);
    if (ErrorCheck(e, "retrieveObject:sm_ReadObject")) return 0;

    return 1;
}

// *************************************************************************
//    Releases an object pinned in the buffer pool.

int Cor_StorageManager::releaseObject(USERDESC *u_desc)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (!SM_initialized) return 0;

    int e = sm_ReleaseObject(u_desc);
    if (ErrorCheck(e, "releaseObject:sm_ReleaseObject")) return 0;

    return 1;
}

// *************************************************************************
//    Destroys an object. If bufGroup is 0, the object is in the catalog file.

int Cor_StorageManager::destroyObject(OID *oid, int bufGroup)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (!SM_initialized) return 0;

    int e;

    // If bufGroup is 0, the object belongs to the catalog
    if (!bufGroup) bufGroup = this->bufGroup();

    e = sm_DestroyObject(bufGroup, oid);
    if (ErrorCheck(e, "destroyObject:sm_DestroyObject")) return 0;

    return 1;
}

// *************************************************************************
int Cor_StorageManager::lookupFile(Cor_CArg& name, int arity, OID *oid)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (!SM_initialized) return 0;

    char buf[COR_MAX_REL_NAME_LEN + 6];

    if (arity > COR_MAX_ARGLIST_SIZE) {
        CORAL_error(COR_TOO_MANY_ARGS, NULL, "lookupFile");
        return 0;
    }

    ASSERT(name.symlen() <= COR_MAX_REL_NAME_LEN);
    sprintf(buf, "%s%3d", name.symbol(), arity);

    return catalog.lookupFile(buf, oid);
}

// *************************************************************************
int Cor_StorageManager::addFile(Cor_CArg& name, int arity,
        Cor_FileDescStruct *obj)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (!SM_initialized)
        return 0;

    char buf[COR_MAX_REL_NAME_LEN + 6];

    if (arity > COR_MAX_ARGLIST_SIZE) {
        CORAL_error(COR_TOO_MANY_ARGS, NULL, "addFile");
        return 0;
    }

    ASSERT(name.symlen() <= COR_MAX_REL_NAME_LEN);
    sprintf(buf, "%s%3d", name.symbol(), arity);

    return catalog.addFile(buf, obj);
}

// *************************************************************************
int Cor_StorageManager::deleteFile(Cor_CArg& name, int arity)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (!SM_initialized) return 0;

    OID oid;
    USERDESC *u_desc;

    // Look up the file in the catalog
    if (! this->lookupFile(name, arity, &oid)) return 0;

    // Retrieve the file descriptor object
    if (! this->retrieveObject(&oid, &u_desc)) return 0;
    Cor_FileDescStruct fd;
    fd.copyDesc((Cor_FileDescStruct *)(u_desc->basePtr));

    // Destroy the file and its contents
    int e = sm_DestroyFile(SM.bufGroup(), &(fd.fid));
    if (ErrorCheck(e, "sm_DestroyFile")) return 0;

    // Destroy all indexes on the file
    for (int i = 0; i < fd.num_indexes; i++) {
        if (fd.indexes[i].valid()) {
            e = sm_DestroyIndex(&(fd.indexes[i].iid), SM.bufGroup());
            if (ErrorCheck(e, "sm_DestroyIndex")) return 0;
        }
    }

    // Delete the entry for the file in the catalogs
    char buf[COR_MAX_REL_NAME_LEN + 6];

    if (arity > COR_MAX_ARGLIST_SIZE) {
        CORAL_error(COR_TOO_MANY_ARGS, NULL, "deleteFile");
        return 0;
    }

    ASSERT(name.symlen() <= COR_MAX_REL_NAME_LEN);
    sprintf(buf, "%s%3d", name.symbol(), arity);

    if (!catalog.deleteFile(buf, &oid)) return 0;
        
    this->releaseObject(u_desc);
    return 1;
}

// *************************************************************************
void Cor_StorageManager::listFiles(ostream *outf)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (!SM_initialized) return;

    catalog.listFiles(outf);
}

// *************************************************************************

#endif WITH_PERSISTENCE
