// *************************************************************************
//    PersistSolvers.C
// *************************************************************************

#include <stdio.h>
#include <iostream.h>
#include <string.h>

#include "CoralStream.h"
#include "BuiltinRelation.h"
#include "PersistentRelation.h"
#include "TupleIterator.h"
#include "Debug.h"
#include "Globals.h"
#include "Util.h"
#include "BuiltinUtil.h"

    // For bulkloading
    // Print something out occasionally so they know we're working on it
#define    NOT_DEAD    10000

    // These items are defined in CArg.y
extern Cor_CArg *Cor_retCArg;
extern Cor_CArg *Cor_RelName;
extern int       Cor_RelArity;

    // Defined in CoralInit.C
extern Cor_DatabaseStruct *Cor_ExodusDB;

// *************************************************************************

#ifdef    WITH_PERSISTENCE

int  Cor_Do_BulkLoad(char *file);
Cor_PersistentRel *Cor_Make_BulkRel(Cor_CArg& relnme, int arity,
        Cor_Tuple *tuple);

#endif   // WITH_PERSISTENCE

extern void deletePersistentRelation(Cor_CArg&, int);

extern int SM_initialized;

// *************************************************************************
static void unavailable(Cor_CArg *cmmd_nme) {

    *(Cor_exEnv->err_strm)
        << "The command '" << *cmmd_nme << "' is unavailable.\n"
        << "Persistence is not compiled into this version of Coral.\n";
}

// *************************************************************************
//  Solver function that is called when the builtin relation schema()
//  is invoked.
// *************************************************************************
Cor_BindEnv *Cor_SchemaSolver(Cor_BuiltinRelation&, Cor_TupleIterator& iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
#ifdef WITH_PERSISTENCE
    int cnt = 0;
    char err_str[100];

    if (!SM_initialized) {
        CORAL_error(COR_ESM_ERR, "dbserver not initialized", "SchemaSolver");
        iterator.set_no_match();
        return iterator.bindenv;
    }

    if (iterator.arg_list.count() > 0) {

        FOR_EACH_ARG(arg, iterator.arg_list) {
            cnt++;
            if (! arg->is_functor()) {
                sprintf(err_str, "bad argument %d - should be a functor", cnt);
                CORAL_error(COR_BAD_ARG_TYPE, err_str, "SchemaSolver");
                iterator.set_no_match();
                return iterator.bindenv;
            }
            
            // Create a file descriptor for the new relation
            int error = 0;
            Cor_FileDescStruct *fd = new Cor_FileDescStruct(arg, error);
            
            // Add the descriptor to the catalogs
            if (!error) {
                SM.addFile(Cor_CArg(arg->functor()),
                    arg->args().arity(), fd);
            } else {
                CORAL_error(error, arg->functor(),"SchemaSolver");
                delete fd;
            }

            // make a new persistent relation
            new Cor_PersistentRel(Cor_CArg(arg->functor()),
                arg->args().arity(), error);
            delete fd;
        } END_EACH_ARG
    }

#else
    unavailable(Cor_SchemaSymbol);
    iterator.set_no_match();
#endif

    return(iterator.bindenv);
}

// **************************************************************************
//
//  Solver function that is called when the builtin relation del_schema()
//  is invoked.
//
// **************************************************************************

Cor_BindEnv *Cor_DelSchemaSolver(Cor_BuiltinRelation&,
        Cor_TupleIterator& iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
#ifdef WITH_PERSISTENCE
    int cnt = 0;
    char err_str[100];

    if (!SM_initialized) {
        CORAL_error(COR_ESM_ERR, "dbserver not initialized", "SchemaSolver");
        iterator.set_no_match();
        return iterator.bindenv;
    }

    if (iterator.arg_list.count() > 0) {

        FOR_EACH_ARG(arg, iterator.arg_list) {
            cnt++;
            if (! arg->is_functor()) {
                sprintf(err_str, "bad argument %d - should be a functor", cnt);
                CORAL_error(COR_BAD_ARG_TYPE, err_str, "SchemaSolver");
                iterator.set_no_match();
                return iterator.bindenv;
            }
            deletePersistentRelation(Cor_CArg(arg->functor()),
                            arg->args().arity());
        } END_EACH_ARG
    }

#else
    unavailable(Cor_DelSchemaSymbol);
    iterator.set_no_match();
#endif
    return(iterator.bindenv);
}

// **************************************************************************
//  Solver function that is called when the builtin relation bulkload()
//  is invoked.
// **************************************************************************

Cor_BindEnv *Cor_BulkLoadSolver(Cor_BuiltinRelation&,
        Cor_TupleIterator& iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
#ifdef WITH_PERSISTENCE

    int j, cnt = iterator.arg_list.count();
    

    if (! SM_initialized) {
        *(Cor_exEnv->err_strm) << 
            "bulkload(). Requires a connection to Exodus.\n";
        iterator.set_no_match(); 
        return(iterator.bindenv);
    }

    if (cnt < 1) {
      *(Cor_exEnv->err_strm) << 
          "bulkload(filename {, filename} ). Supply files as arguments.\n";
      iterator.set_no_match(); 
    } else {

      for (j=0; j < cnt; j++) {

          Cor_Term term (iterator.arg_list[j], iterator.bindenv);
          Cor_deep_dereference_term (term);

          if (! term.expr.is_symbol()) {
              *(Cor_exEnv->err_strm) << 
            "bulkload(filename {, filename} ). Supply files as arguments.\n";
              iterator.set_no_match(); 
              return(iterator.bindenv);
          }

          char *filename = term.expr.symbol();
          if (! Cor_Do_BulkLoad(filename)) {
              *(Cor_exEnv->err_strm)
                   << "Bulk-Loader returned an error for file: "
                   << filename << endl; 
          }
      }
    }

#else
    unavailable(Cor_BulkLoadSymbol);
    iterator.set_no_match();
#endif
    return(iterator.bindenv);
}

// **************************************************************************
//  Solver function that is called when the builtin relation StartBulkLoad()
//  is invoked.
// **************************************************************************

Cor_BindEnv *Cor_StartBulkLoadSolver(Cor_BuiltinRelation&,
        Cor_TupleIterator& iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    *(Cor_exEnv->err_strm)
        << "The builtin 'start_bulk_load' is no longer implemented.\n"
        << "Use 'bulkload(filename)' to bulk-load facts into Exodus.\n";

    iterator.set_no_match();
    return(iterator.bindenv);
}

// **************************************************************************
//  Solver function that is called when the builtin relation EndBulkLoad()
//  is invoked.
// **************************************************************************

Cor_BindEnv *Cor_EndBulkLoadSolver(Cor_BuiltinRelation&,
            Cor_TupleIterator& iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    *(Cor_exEnv->err_strm)
        << "The builtin 'end_bulk_load' is no longer implemented.\n"
        << "Use 'bulkload(filename)' to bulk-load facts into Exodus.\n";

    iterator.set_no_match();
    return(iterator.bindenv);
}

// **************************************************************************
Cor_BindEnv *Cor_TransBeginSolver(Cor_BuiltinRelation&,
        Cor_TupleIterator& iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
#ifdef WITH_PERSISTENCE

    if (!SM_initialized) {
        iterator.set_no_match();
        return NULL;
    }

    SM.startTrans();
    if (!SM_initialized) {
        iterator.set_no_match();
        return NULL;
    }
 
#else
    unavailable(Cor_BeginTransSymbol);
    iterator.set_no_match();
#endif

    return(iterator.bindenv);
}

// **************************************************************************
Cor_BindEnv *Cor_TransAbortSolver(Cor_BuiltinRelation&,
        Cor_TupleIterator& iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
#ifdef WITH_PERSISTENCE

    if (!SM_initialized) {
        iterator.set_no_match();
        return NULL;
    }

    SM.abortTrans();

    if (!SM_initialized) {
        iterator.set_no_match();
        return NULL;
    }
 
#else
    unavailable(Cor_AbortTransSymbol);
    iterator.set_no_match();
#endif

    return(iterator.bindenv);
}

// **************************************************************************
Cor_BindEnv *Cor_TransCommitSolver(Cor_BuiltinRelation&,
            Cor_TupleIterator& iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
#ifdef WITH_PERSISTENCE

    if (!SM_initialized) {
        iterator.set_no_match();
        return NULL;
    }

    if (!SM.commitTrans() || !SM_initialized) {
        iterator.set_no_match();
        return NULL;
    }
 
#else
    unavailable(Cor_CommitTransSymbol);
    iterator.set_no_match();
#endif
    return iterator.bindenv;
}

// *************************************************************************
#ifdef   WITH_PERSISTENCE

int Cor_Do_BulkLoad( char *file )
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (Cor_retCArg) {
        Cor_retCArg->destroy();
        Cor_retCArg = NULL;
    } 

    if (Cor_RelName) {
        Cor_RelName->destroy();
        Cor_RelName = NULL;
    } 

      // Now consult the fact or table file 
    Cor_consultFile(file,file);

      // Start up a transaction
    SM.startTrans();

    // Now read in the lines using parseFact() and insert into the relation

    int cnt = 0;
    Cor_PersistentRel *rel = NULL;
    Cor_Tuple  *tuple = NULL;

    Cor_CArg prev_relnme;
    Cor_CArg relnme;
    int rel_arity = 0;
    int prev_arity = 0;

    tuple = Cor_parseFact(NULL, relnme, &rel_arity);

    //(*cerr_strm) << *tuple;  // Bus error
    //(*cerr_strm) << endl;
    
    while (tuple) {

            // If its a new relation find or create it
        if ((rel_arity != prev_arity) || (! relnme.equals(prev_relnme)) ) {

            rel = (Cor_PersistentRel *)
                        Cor_ExodusDB->RelationTable->lookup(relnme);

            //rel = Cor_find_local_or_external_relation(relnme, rel_arity);
            if (! rel) {

                // Need to make it
               rel = Cor_Make_BulkRel(relnme, rel_arity, tuple);
               if (!rel) {
                  *(Cor_exEnv->err_strm) << "Error creating relation ";
                  *(Cor_exEnv->err_strm) << relnme << endl;
                           //  << relnme << endl;
                  break;
               }
            }

            prev_relnme = relnme;
            prev_arity  = rel_arity;
        }

        
          // Insert fact into Relation rel
        //tuple->printon(cerr_strm);
        //(*cerr_strm) << endl;

        if (rel)
            rel->insert_raw(tuple);

          // Can delete everything, the data is in Exodus now
        delete tuple->_args;
        delete tuple;

        if ((++cnt % NOT_DEAD) == 0) {
            *(Cor_exEnv->out_strm) << cnt << endl;

            // Will this reduce the Exodus log space required?
            SM.commitTrans();
            SM.startTrans();
        }

        tuple = Cor_parseFact(NULL, relnme, &rel_arity);
    }

    SM.commitTrans();
    return(1);
}

// *************************************************************************
//  Tuple is an example of a fact
Cor_PersistentRel *Cor_Make_BulkRel(Cor_CArg& relnme, int arity,
        Cor_Tuple *tuple)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int error = 0;

      // Create a file descriptor for the new relation

    Cor_CArg *f_arg = new Cor_CArg(relnme,arity);
    for (int j=0; j < arity; j++) {
        (f_arg->args())[j] = (*(tuple->_args))[j];
    }

    Cor_FileDescStruct *fd = new Cor_FileDescStruct(f_arg, error, 0, 1);

    // SPF - need something here, or just delete it?
    //f_arg->destroy();

    // Add the descriptor to the catalogs
    if (!error) {
        SM.addFile(relnme, arity, fd);
    } else {
        *(Cor_exEnv->err_strm) << "Error: could not open an Exodus file.\n";
        delete fd;
        return(NULL);
    }

      // make a new persistent relation
    Cor_PersistentRel *rel =
            new Cor_PersistentRel(relnme, arity, error);

    delete fd;
    return(rel);
}
#endif   // WITH_PERSISTENCE

// *************************************************************************
